# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from tqdm.auto import trange

from ..recommender import Recommender
from ...exception import ScoreException
from ...utils import get_rng
from ...utils.init_utils import normal


class HFT(Recommender):
    """Hidden Factors and Hidden Topics

    Parameters
    ----------
    name: string, default: 'HFT'
        The name of the recommender model.

    k: int, optional, default: 10
        The dimension of the latent factors.

    max_iter: int, optional, default: 50
        Maximum number of iterations for EM.

    grad_iter: int, optional, default: 50
        Maximum number of iterations for L-BFGS.

    lambda_text: float, default: 0.1
        Weight of corpus likelihood in objective function.

    l2_reg: float, default: 0.001
        Regularization for user item latent factors.

    vocab_size: int, optional, default: 8000
        Size of vocabulary for review text.

    init_params: dictionary, optional, default: None
        List of initial parameters, e.g., init_params = {'alpha': alpha, 'beta_u': beta_u,
        'beta_i': beta_i, 'gamma_u': gamma_u, 'gamma_v': gamma_v}

        alpha: float
            Model offset, optional initialization via init_params.

        beta_u: ndarray. shape (n_user, 1)
            User biases, optional initialization via init_params.

        beta_u: ndarray. shape (n_item, 1)
            Item biases, optional initialization via init_params.

        gamma_u: ndarray, shape (n_users,k)
            The user latent factors, optional initialization via init_params.

        gamma_v: ndarray, shape (n_items,k)
            The item latent factors, optional initialization via init_params.

    trainable: boolean, optional, default: True
        When False, the model will not be re-trained, and input of pre-trained parameters are required.

    verbose: boolean, optional, default: True
        When True, some running logs are displayed.
        
    seed: int, optional, default: None
        Random seed for weight initialization.

    References
    ----------
    Julian McAuley, Jure Leskovec. "Hidden Factors and Hidden Topics: Understanding Rating Dimensions with Review Text"
    RecSys '13 Proceedings of the 7th ACM conference on Recommender systems Pages 165-172
    """

    def __init__(
        self,
        name="HFT",
        k=10,
        max_iter=50,
        grad_iter=50,
        lambda_text=0.1,
        l2_reg=0.001,
        vocab_size=8000,
        init_params=None,
        trainable=True,
        verbose=True,
        seed=None,
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)

        self.k = k
        self.lambda_text = lambda_text
        self.l2_reg = l2_reg
        self.grad_iter = grad_iter
        self.name = name
        self.max_iter = max_iter
        self.verbose = verbose
        self.seed = seed
        self.vocab_size = vocab_size

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.alpha = self.init_params.get("alpha", None)
        self.beta_u = self.init_params.get("beta_u", None)
        self.beta_i = self.init_params.get("beta_i", None)
        self.gamma_u = self.init_params.get("gamma_u", None)
        self.gamma_i = self.init_params.get("gamma_i", None)

    def _init(self):
        rng = get_rng(self.seed)
        self.n_item = self.train_set.num_items
        self.n_user = self.train_set.num_users

        if self.alpha is None:
            self.alpha = self.train_set.global_mean
        if self.beta_u is None:
            self.beta_u = normal(self.n_user, std=0.01, random_state=rng)
        if self.beta_i is None:
            self.beta_i = normal(self.n_item, std=0.01, random_state=rng)
        if self.gamma_u is None:
            self.gamma_u = normal((self.n_user, self.k), std=0.01, random_state=rng)
        if self.gamma_i is None:
            self.gamma_i = normal((self.n_item, self.k), std=0.01, random_state=rng)

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        self._init()

        if self.trainable:
            self._fit_hft()

        return self

    @staticmethod
    def _build_data(csr_mat):
        index_list = []
        rating_list = []
        for i in range(csr_mat.shape[0]):
            j, k = csr_mat.indptr[i], csr_mat.indptr[i + 1]
            index_list.append(csr_mat.indices[j:k])
            rating_list.append(csr_mat.data[j:k])
        return index_list, rating_list

    def _fit_hft(self):
        from .hft import Model

        # document data
        bow_mat = self.train_set.item_text.batch_bow(
            np.arange(self.n_item), keep_sparse=True
        )
        documents, _ = self._build_data(bow_mat)  # bag of word feature
        # Rating data
        user_data = self._build_data(self.train_set.matrix)
        item_data = self._build_data(self.train_set.matrix.T.tocsr())

        model = Model(
            n_user=self.n_user,
            n_item=self.n_item,
            alpha=self.alpha,
            beta_u=self.beta_u,
            beta_i=self.beta_i,
            gamma_u=self.gamma_u,
            gamma_i=self.gamma_i,
            n_vocab=self.vocab_size,
            k=self.k,
            lambda_text=self.lambda_text,
            l2_reg=self.l2_reg,
            grad_iter=self.grad_iter,
        )

        model.init_count(docs=documents)

        # training
        loop = trange(self.max_iter, disable=not self.verbose)
        for _ in loop:
            model.assign_word_topics(docs=documents)
            loss = model.update_params(rating_data=(user_data, item_data))
            loop.set_postfix(loss=loss)

        self.alpha, self.beta_u, self.beta_i, self.gamma_u, self.gamma_i = (
            model.get_parameter()
        )

        if self.verbose:
            print("Learning completed!")

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items
        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            known_item_scores = (
                self.alpha
                + self.beta_u[user_idx]
                + self.beta_i
                + self.gamma_i.dot(self.gamma_u[user_idx, :])
            )
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )

            user_pred = (
                self.alpha
                + self.beta_u[user_idx]
                + self.beta_i[item_idx]
                + self.gamma_i[item_idx, :].dot(self.gamma_u[user_idx, :])
            )

            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

# cython: language_level=3

from libc.math cimport exp
from libc.math cimport sqrt

import numpy as np
from ...utils.init_utils import normal
from ...utils import get_rng


#Sigmoid function
cdef float sigmoid(float z):
    cdef float s
    cdef float MAX_EXP = 6.0
    if z > MAX_EXP:
        return 1.0
    else:
        if z < -MAX_EXP:
            return 0.0
        else:
            s = 1.0 / (1.0 + exp(-z))
            return s


def sorec(int[:] rat_uid, int[:] rat_iid, float[:] rat_val, int[:] net_uid, int[:] net_jid, float[:] net_val,int k,
        int n_users, int n_items, int n_ratings, int n_edges, int n_epochs = 100, float lamda_c = 10, float lamda = 0.001,
        float learning_rate = 0.001, float gamma = 0.9, init_params = None, verbose = False, seed=None):

    cdef:
        double[:] loss = np.full(n_epochs, 0.0)
        int n = n_users
        int d = n_items
        double[:,:] U
        double[:,:] V
        double[:,:] Z
        double[:,:] cache_u = np.zeros((n,k))
        double[:,:] cache_v = np.zeros((d,k))
        double[:,:] cache_z = np.zeros((n,k))
        double[:,:] grad_u = np.zeros((n,k))
        double[:,:] grad_v = np.zeros((d,k))
        double[:,:] grad_z = np.zeros((n,k))
        double eps = 1e-8
        int u_, i_, j_, k_, r, ed, epoch
        double val, s, e, norm_u, norm_v


    # Initialize factors
    rng = get_rng(seed)

    U = init_params.get('U', None)
    if U is None:
        U = normal((n, k), mean=0.0, std=0.001, random_state=rng, dtype=np.double)
    
    V = init_params.get('V', None)
    if V is None:
        V = normal((d, k), mean=0.0, std=0.001, random_state=rng, dtype=np.double)
    
    Z = init_params.get('Z', None)
    if Z is None:
        Z = normal((n, k), mean=0.0, std=0.001, random_state=rng, dtype=np.double)


    # Optimization
    for epoch in range(n_epochs):

        for ed in range(n_edges):
            i_, j_, val = net_uid[ed], net_jid[ed], net_val[ed]

            s = 0.0
            for k_ in range(k):
                s += U[i_, k_] * Z[j_, k_]
            sg = sigmoid(s)
            err = (val - sg)  # Error for the obseved rating i_, j_, val
            werr = err * sg * (1. - sg)  # Weighted error

            # update user factors
            for k_ in range(k):
                grad_u[i_, k_] = werr * Z[j_, k_] - lamda * U[i_, k_]
                cache_u[i_, k_] = gamma * cache_u[i_, k_] + (1 - gamma) * (grad_u[i_, k_] * grad_u[i_, k_])
                U[i_, k_] += lamda_c * learning_rate * (grad_u[i_, k_] / (sqrt(cache_u[i_, k_]) + eps))

            # update social network factors
            for k_ in range(k):
                grad_z[j_, k_] = werr * U[i_, k_] - lamda * Z[j_, k_]
                cache_z[j_, k_] = gamma * cache_z[j_, k_] + (1 - gamma) * (grad_z[j_, k_] * grad_z[j_, k_])
                Z[j_, k_] += lamda_c * learning_rate * (grad_z[j_, k_] / (sqrt(cache_z[j_, k_]) + eps))

            norm_z = 0.0
            norm_u = 0.0
            for k_ in range(k):
                norm_z += Z[j_, k_] * Z[j_, k_]
                norm_u += U[i_, k_] * U[i_, k_]

            loss[epoch] += err * err + lamda * (norm_z + norm_u)

        for r in range(n_ratings):
            u_, i_, val = rat_uid[r], rat_iid[r], rat_val[r]

            s = 0.0
            for k_ in range(k):
                s += U[u_, k_] * V[i_, k_]
            sg = sigmoid(s)
            err = (val - sg)  # Error for the obseved rating u_, i_, val
            werr = err * sg * (1. - sg)  # Weighted error

            # update user factors
            for k_ in range(k):
                grad_u[u_, k_] = werr * V[i_, k_] - lamda * U[u_, k_]
                cache_u[u_, k_] = gamma * cache_u[u_, k_] + (1 - gamma) * (grad_u[u_, k_] * grad_u[u_, k_])
                U[u_, k_] += learning_rate * (grad_u[u_, k_] / (sqrt(cache_u[
                                                                     u_, k_]) + eps))  # Update the user factor, better to reweight the L2 regularization terms acoording the number of ratings per-user

            # update item factors
            for k_ in range(k):
                grad_v[i_, k_] = werr * U[u_, k_] - lamda * V[i_, k_]
                cache_v[i_, k_] = gamma * cache_v[i_, k_] + (1 - gamma) * (grad_v[i_, k_] * grad_v[i_, k_])
                V[i_, k_] += learning_rate * (grad_v[i_, k_] / (sqrt(cache_v[i_, k_]) + eps))

            norm_u = 0.0
            norm_v = 0.0
            for k_ in range(k):
                norm_u += U[u_, k_] * U[u_, k_]
                norm_v += V[i_, k_] * V[i_, k_]

            loss[epoch] += err * err + lamda * (norm_u + norm_v)

        if verbose:
            print('epoch %i, loss: %f' % (epoch, loss[epoch]))

    res = {'U': U, 'V': V, 'Z': Z, 'loss': loss}

    return res



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import tensorflow as tf


def conv_layer(
    input,
    num_input_channels,
    filter_height,
    filter_width,
    num_filters,
    seed=None,
    use_pooling=True,
):
    shape = [filter_height, filter_width, num_input_channels, num_filters]
    weights = tf.Variable(tf.truncated_normal(shape, stddev=0.05, seed=seed))
    biases = tf.Variable(tf.constant(0.05, shape=[num_filters]))
    layer = tf.nn.conv2d(
        input=input, filter=weights, strides=[1, 1, 1, 1], padding="VALID"
    )
    layer = layer + biases
    if use_pooling:
        layer = tf.nn.max_pool(
            value=layer,
            ksize=[1, input.shape[1] - filter_height + 1, 1, 1],
            strides=[1, 1, 1, 1],
            padding="VALID",
        )
    layer = tf.nn.relu(layer)
    return layer, weights


def flatten_layer(layer):
    layer_shape = layer.get_shape()
    num_feature = layer_shape[1:4].num_elements()
    layer_flat = tf.reshape(layer, [-1, num_feature])
    return layer_flat, num_feature


def fc_layer(input, num_input, num_output, seed=None):
    weights = tf.Variable(
        tf.truncated_normal([num_input, num_output], stddev=0.05, seed=seed)
    )
    biases = tf.Variable(tf.constant(0.05, shape=[num_output]))
    layer = tf.matmul(input, weights) + biases
    layer = tf.nn.tanh(layer)
    return layer


class CNN_module:
    def __init__(
        self,
        output_dimension,
        dropout_rate,
        emb_dim,
        max_len,
        filter_sizes,
        num_filters,
        hidden_dim,
        seed,
        init_W,
        learning_rate=0.001,
    ):
        self.drop_rate = dropout_rate
        self.max_len = max_len
        self.seed = seed
        self.learning_rate = learning_rate
        self.init_W = tf.constant(init_W)
        self.output_dimension = output_dimension
        self.emb_dim = emb_dim
        self.filter_lengths = filter_sizes
        self.nb_filters = num_filters
        self.vanila_dimension = hidden_dim

        self._build_graph()

    def _build_graph(self):
        # create Graph
        self.model_input = tf.placeholder(dtype=tf.int32, shape=(None, self.max_len))
        self.v = tf.placeholder(dtype=tf.float32, shape=(None, self.output_dimension))
        self.sample_weight = tf.placeholder(dtype=tf.float32, shape=(None,))
        self.embedding_weight = tf.Variable(initial_value=self.init_W)

        self.seq_emb = tf.nn.embedding_lookup(self.embedding_weight, self.model_input)
        self.reshape = tf.reshape(self.seq_emb, [-1, self.max_len, self.emb_dim, 1])
        self.convs = []

        # Convolutional layers
        for i in self.filter_lengths:
            convolutional_layer, weights = conv_layer(
                input=self.reshape,
                num_input_channels=1,
                filter_height=i,
                filter_width=self.emb_dim,
                num_filters=self.nb_filters,
                use_pooling=True,
            )

            flat_layer, _ = flatten_layer(convolutional_layer)
            self.convs.append(flat_layer)

        self.model_output = tf.concat(self.convs, axis=-1)
        # Fully-connected layers
        self.model_output = fc_layer(
            input=self.model_output,
            num_input=self.model_output.get_shape()[-1].value,
            num_output=self.vanila_dimension,
        )
        # Dropout layer
        self.model_output = tf.nn.dropout(self.model_output, self.drop_rate)
        # Output layer
        self.model_output = fc_layer(
            input=self.model_output,
            num_input=self.vanila_dimension,
            num_output=self.output_dimension,
        )
        # Weighted MEA loss function
        self.mean_square_loss = tf.losses.mean_squared_error(
            labels=self.v,
            predictions=self.model_output,
            reduction=tf.losses.Reduction.NONE,
        )
        self.weighted_loss = tf.reduce_sum(
            tf.reduce_sum(self.mean_square_loss, axis=1, keepdims=True)
            * self.sample_weight
        )
        # RMSPro optimizer
        self.optimizer = tf.train.RMSPropOptimizer(
            learning_rate=self.learning_rate
        ).minimize(self.weighted_loss)



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example to run Non-negative Matrix Factorization (NMF) model with Ratio Split evaluation strategy"""

import cornac
from cornac.datasets import movielens
from cornac.eval_methods import RatioSplit

# Load the MovieLens 100K dataset
ml_100k = movielens.load_feedback()

# Instantiate an evaluation method.
eval_method = RatioSplit(
    data=ml_100k,
    test_size=0.2,
    rating_threshold=4.0,
    exclude_unknowns=True,
    verbose=True,
    seed=123,
)

# Instantiate a NMF recommender model.
nmf = cornac.models.NMF(
    k=15,
    max_iter=50,
    learning_rate=0.005,
    lambda_u=0.06,
    lambda_v=0.06,
    lambda_bu=0.02,
    lambda_bi=0.02,
    use_bias=False,
    verbose=True,
    seed=123,
)

# Instantiate evaluation metrics.
mae = cornac.metrics.MAE()
rmse = cornac.metrics.RMSE()
rec_20 = cornac.metrics.Recall(k=20)
pre_20 = cornac.metrics.Precision(k=20)

# Instantiate and then run an experiment.
cornac.Experiment(
    eval_method=eval_method,
    models=[nmf],
    metrics=[mae, rmse, rec_20, pre_20],
    user_based=True,
).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================


import tensorflow as tf


act_functions = {
    "sigmoid": tf.nn.sigmoid,
    "tanh": tf.nn.tanh,
    "elu": tf.nn.elu,
    "selu": tf.nn.selu,
    "relu": tf.nn.relu,
    "relu6": tf.nn.relu6,
    "leaky_relu": tf.nn.leaky_relu,
}


def loss_fn(labels, logits):
    cross_entropy = tf.reduce_mean(
        tf.nn.sigmoid_cross_entropy_with_logits(labels=labels, logits=logits)
    )
    reg_loss = tf.losses.get_regularization_loss()
    return cross_entropy + reg_loss


def train_fn(loss, learning_rate, learner):
    if learner.lower() == "adagrad":
        opt = tf.train.AdagradOptimizer(learning_rate=learning_rate, name="optimizer")
    elif learner.lower() == "rmsprop":
        opt = tf.train.RMSPropOptimizer(learning_rate=learning_rate, name="optimizer")
    elif learner.lower() == "adam":
        opt = tf.train.AdamOptimizer(learning_rate=learning_rate, name="optimizer")
    else:
        opt = tf.train.GradientDescentOptimizer(
            learning_rate=learning_rate, name="optimizer"
        )

    return opt.minimize(loss)


def emb(
    uid, iid, num_users, num_items, emb_size, reg_user, reg_item, seed=None, scope="emb"
):
    with tf.variable_scope(scope):
        user_emb = tf.get_variable(
            "user_emb",
            shape=[num_users, emb_size],
            dtype=tf.float32,
            initializer=tf.random_normal_initializer(stddev=0.01, seed=seed),
            regularizer=tf.contrib.layers.l2_regularizer(scale=reg_user),
        )
        item_emb = tf.get_variable(
            "item_emb",
            shape=[num_items, emb_size],
            dtype=tf.float32,
            initializer=tf.random_normal_initializer(stddev=0.01, seed=seed),
            regularizer=tf.contrib.layers.l2_regularizer(scale=reg_item),
        )

    return tf.nn.embedding_lookup(user_emb, uid), tf.nn.embedding_lookup(item_emb, iid)


def gmf(uid, iid, num_users, num_items, emb_size, reg_user, reg_item, seed=None):
    with tf.variable_scope("GMF") as scope:
        user_emb, item_emb = emb(
            uid=uid,
            iid=iid,
            num_users=num_users,
            num_items=num_items,
            emb_size=emb_size,
            reg_user=reg_user,
            reg_item=reg_item,
            seed=seed,
            scope=scope,
        )
        return tf.multiply(user_emb, item_emb)


def mlp(uid, iid, num_users, num_items, layers, reg_layers, act_fn, seed=None):
    with tf.variable_scope("MLP") as scope:
        user_emb, item_emb = emb(
            uid=uid,
            iid=iid,
            num_users=num_users,
            num_items=num_items,
            emb_size=layers[0] / 2,
            reg_user=reg_layers[0],
            reg_item=reg_layers[0],
            seed=seed,
            scope=scope,
        )
        interaction = tf.concat([user_emb, item_emb], axis=-1)
        for i, layer in enumerate(layers[1:]):
            interaction = tf.layers.dense(
                interaction,
                units=layer,
                name="layer{}".format(i + 1),
                activation=act_functions.get(act_fn, tf.nn.relu),
                kernel_initializer=tf.initializers.lecun_uniform(seed),
                kernel_regularizer=tf.contrib.layers.l2_regularizer(reg_layers[i + 1]),
            )
        return interaction





# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example SKMeans vs BPR on MovieLens data"""

import cornac
from cornac.datasets import movielens
from cornac.eval_methods import RatioSplit


# Load user-item feedback
data = movielens.load_feedback(variant="100K")

# Instantiate an evaluation method to split data into train and test sets.
ratio_split = RatioSplit(
    data=data,
    test_size=0.2,
    exclude_unknowns=True,
    verbose=True,
    seed=123,
    rating_threshold=0.5,
)

# Instantiate models
skm = cornac.models.SKMeans(
   k=5,
   max_iter=100,
   tol=1e-10,
   seed=123
)

bpr = cornac.models.BPR(
   k=5,
   max_iter=200,
   learning_rate=0.001,
   lambda_reg=0.01,
   seed=123)


# Instantiate evaluation measures
rec_20 = cornac.metrics.Recall(k=20)
ndcg_20 = cornac.metrics.NDCG(k=20)
auc = cornac.metrics.AUC()



# Put everything together into an experiment and run it
cornac.Experiment(
    eval_method=ratio_split,
    models=[skm, bpr],
    metrics=[rec_20, ndcg_20, auc],
    user_based=True,
).run()


# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example for Hidden Factors as Topics (HFT) with Movilen 1M dataset """

import cornac
from cornac.data import Reader
from cornac.datasets import movielens
from cornac.eval_methods import RatioSplit
from cornac.data import TextModality
from cornac.data.text import BaseTokenizer


# HFT jointly models the user-item preferences and item texts (e.g., product reviews) with shared item factors
# Below we fit HFT to the MovieLens 1M dataset. We need  both the ratings and movie plots information
plots, movie_ids = movielens.load_plot()
ml_1m = movielens.load_feedback(variant="1M", reader=Reader(item_set=movie_ids))

# Instantiate a TextModality, it makes it convenient to work with text auxiliary information
# For more details, please refer to the tutorial on how to work with auxiliary data
item_text_modality = TextModality(
    corpus=plots,
    ids=movie_ids,
    tokenizer=BaseTokenizer(sep="\t", stop_words="english"),
    max_vocab=5000,
    max_doc_freq=0.5,
)

# Define an evaluation method to split feedback into train and test sets
ratio_split = RatioSplit(
    data=ml_1m,
    test_size=0.2,
    exclude_unknowns=True,
    item_text=item_text_modality,
    verbose=True,
    seed=123,
)

# Instantiate HFT model
hft = cornac.models.HFT(
    k=10,
    max_iter=40,
    grad_iter=5,
    l2_reg=0.001,
    lambda_text=0.01,
    vocab_size=5000,
    seed=123,
)

# Instantiate MSE for evaluation
mse = cornac.metrics.MSE()

# Put everything together into an experiment and run it
cornac.Experiment(
    eval_method=ratio_split, models=[hft], metrics=[mse], user_based=False
).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

# cython: language_level=3

import multiprocessing

cimport cython
from cython.operator cimport dereference
from cython.parallel import parallel, prange
from libc.stdlib cimport abort, malloc, free
from libcpp cimport bool
from libcpp.string cimport string

import numpy as np
cimport numpy as np
from tqdm.auto import trange

from ..recommender import Recommender
from ...exception import ScoreException
from ...utils import get_rng
from ...utils import fast_dot
from ...utils.common import scale
from ...utils.init_utils import zeros, normal

from .recom_fm cimport Data
from .recom_fm cimport DataMetaInfo
from .recom_fm cimport DVectorDouble
from .recom_fm cimport sparse_entry
from .recom_fm cimport sparse_row
from .recom_fm cimport LargeSparseMatrixMemory

from .recom_fm cimport fm_model
from .recom_fm cimport fm_learn
from .recom_fm cimport fm_learn_sgd
from .recom_fm cimport fm_learn_sgd_element
from .recom_fm cimport fm_learn_sgd_element_adapt_reg
from .recom_fm cimport fm_learn_mcmc
from .recom_fm cimport fm_learn_mcmc_simultaneous



cdef Data* _prepare_data(long[:] uid, long[:] iid, float[:] val, long num_feature, bool has_t, bool verbose):
        cdef num_values = 2 * len(val)
        cdef num_rows = len(val)

        cdef Data *data = new Data(0, False, has_t)
        data.verbose = verbose

        cdef LargeSparseMatrixMemory[float] *X = new LargeSparseMatrixMemory[float]()
        cdef DVector[float] *target = new DVector[float]()

        X.data.setSize(num_rows)
        target.setSize(num_rows)

        cdef sparse_entry[float] *cache = <sparse_entry[float] *> malloc(num_values * sizeof(sparse_entry[float]))
        if cache is NULL:
            abort()

        cdef unsigned int row_id
        cdef unsigned int cache_id = 0

        for row_id in range(num_rows):
            target.set(row_id, val[row_id])
            X.data.value[row_id].data = &cache[cache_id]
            X.data.value[row_id].size = 0

            # user feature
            cache[cache_id].id = uid[row_id]
            cache[cache_id].value = 1
            cache_id += 1
            X.data.value[row_id].size += 1

            # item feature
            cache[cache_id].id = iid[row_id]
            cache[cache_id].value = 1
            cache_id += 1
            X.data.value[row_id].size += 1 

        assert num_values == cache_id

        X.num_values = num_values
        X.num_cols = num_feature

        data.load(dereference(X), dereference(target))

        return data



class FM(Recommender):
    """Factorization Machines.

    Parameters
    ----------
    k0: int, optional, default: 1
        Using bias 'w0'.
    
    k1: int, optional, default: 1
        Using first-order weights 'w'.

    k2: int, optional, default: 8
        Dimension of second-order weights 'v'.

    max_iter: int, optional, default: 100
        Maximum number of iterations or the number of epochs for SGD, SGDA.

    learning_rate: float, optional, default: 0.01
        The learning rate for SGD.

    reg0: float, optional, default: 0.0
        Bias regularization. 

    reg1: float, optional, default: 0.0
        First-order weights regularization. 

    reg2: float, optional, default: 0.0
        Second-order weights regularization. 

    reg_all: float, optional, default: 0.0
        Regularization for all parameters.
        If 'reg_all' != 0.0, this will be used for all regularization.

    init_stdev: float, optional, default: 0.1
        Standard deviation for initialization of second-order weights 'v'.

    method, str, optional, default: 'mcmc'
        Learning method (SGD, SGDA, ALS, MCMC)

    num_threads: int, optional, default: 0
        Number of parallel threads for training. If num_threads=0, all CPU cores will be utilized.
        If seed is not None, num_threads=1 to remove randomness from parallelization.

    trainable: boolean, optional, default: True
        When False, the model will not be re-trained, and input of pre-trained parameters are required.

    verbose: boolean, optional, default: True
        When True, some running logs are displayed.

    init_params: dictionary, optional, default: None
        Initial parameters, e.g., init_params = {'w0': bias, 'w': first-order, 'v': second-order}

    seed: int, optional, default: None
        Random seed for weight initialization.
        If specified, training will take longer because of single-thread (no parallelization).

    References
    ----------
    * Rendle, S. (2010, December). Factorization machines. \
    In 2010 IEEE International Conference on Data Mining (pp. 995-1000). IEEE.
    
    * Rendle, S. (2012). Factorization machines with libfm. \
    ACM Transactions on Intelligent Systems and Technology (TIST), 3(3), 1-22.
    """

    def __init__(
        self, 
        name="FM",
        k0=1,
        k1=1,
        k2=8,
        max_iter=100,
        learning_rate=0.01,
        reg0=0.0,
        reg1=0.0,
        reg2=0.0,
        reg_all=0.0,
        init_stdev=0.1,
        method="mcmc",
        num_threads=0, 
        trainable=True, 
        verbose=True, 
        init_params=None, 
        seed=None
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)
        self.k0 = k0
        self.k1 = k1
        self.k2 = k2
        self.max_iter = max_iter
        self.learning_rate = learning_rate

        self.reg0 = reg_all if reg_all != 0.0 else reg0
        self.reg1 = reg_all if reg_all != 0.0 else reg1
        self.reg2 = reg_all if reg_all != 0.0 else reg2
        
        self.init_stdev = init_stdev
        self.method = method.lower()
        self.seed = seed
        self.rng = get_rng(seed)

        if seed is not None:
            self.num_threads = 1
        elif num_threads > 0 and num_threads < multiprocessing.cpu_count():
            self.num_threads = num_threads
        else:
            self.num_threads = multiprocessing.cpu_count()

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.w0 = self.init_params.get('w0', None)
        self.w = self.init_params.get('w', None)
        self.v = self.init_params.get('v', None)
    
    def _init(self):
        num_features = self.train_set.total_users + self.train_set.total_items

        if self.w0 is None:
            self.w0 = 0.0
        if self.w is None:
            self.w = zeros(num_features, dtype=np.double)
        if self.v is None:
            d = self.k2 if self.k2 else 1  # dummy if self.k2 == 0
            self.v = normal((d, num_features), std=self.init_stdev, random_state=self.seed, dtype=np.double)


    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        self._init()

        if not self.trainable:
            return self

        self._fit_libfm(train_set, val_set, self.w, self.v)

        if self.verbose:
            print('Optimization finished!')

        return self


    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _fit_libfm(self, train_set, val_set, double[:] w, double[:, :] v):
        cdef unsigned int num_feature = self.train_set.total_users + self.train_set.total_items

        (uid, iid, val) = self.train_set.uir_tuple
        cdef Data *train = _prepare_data(
            uid, 
            iid + self.train_set.total_users, 
            val.astype(np.float32), 
            num_feature, 
            self.method in ["als", "mcmc"],
            self.verbose,
        )
        if self.verbose:
            print("Training data:")
            train.debug()

        cdef Data *validation = NULL
        if val_set is not None:
            (uid, iid, val) = val_set.uir_tuple
            validation = _prepare_data(
                uid, 
                iid + self.train_set.total_users,
                val.astype(np.float32), 
                num_feature, 
                self.method in ["als", "mcmc"],
                self.verbose,
            )
            if self.verbose:
                print("Validation data:")
                validation.debug()
            
        # (1.3) Load meta data
        # (main table)
        cdef DataMetaInfo *meta = new DataMetaInfo(num_feature)
        
        # meta.num_attr_groups = meta_main.num_attr_groups
        meta.num_attr_per_group.setSize(meta.num_attr_groups)
        meta.num_attr_per_group.init(0)
        cdef unsigned int i, j
        for i in range(meta.attr_group.dim):
            j = meta.attr_group.get(i)
            meta.num_attr_per_group.set(j, meta.num_attr_per_group.get(j) + 1)
        
        if self.verbose:
            meta.debug()

        # (2) Setup the factorization machine
        cdef fm_model fm

        fm.num_attribute = num_feature
        fm.init_stdev = self.init_stdev
        # set the number of dimensions in the factorization
        fm.k0 = self.k0 != 0
        fm.k1 = self.k1 != 0
        fm.num_factor = self.k2
        fm.init()
        # reset the weights of v
        if self.k2:    
            for i in range(self.k2):
                for j in range(num_feature):
                    fm.v.set(i, j, v[i, j]) 

        # (3) Setup the learning method:
        cdef fm_learn *fml

        if self.method == "sgd":
            fml = new fm_learn_sgd_element()
            (<fm_learn_sgd*>fml).num_iter = self.max_iter
        elif self.method == "sgda":
            if val_set is None:
                raise ValueError("'sgda' method requires validation set but None, other methods: ['sgd', 'als', 'mcmc']" )
            
            fml = new fm_learn_sgd_element_adapt_reg()
            (<fm_learn_sgd*>fml).num_iter = self.max_iter
            (<fm_learn_sgd_element_adapt_reg*>fml).validation = validation
        else: # als or mcmc
            fm.w.init_normal(fm.init_mean, fm.init_stdev)
            fml = new fm_learn_mcmc_simultaneous()
            (<fm_learn_mcmc*>fml).num_iter = self.max_iter
            (<fm_learn_mcmc*>fml).do_sample = self.method == "mcmc"
            (<fm_learn_mcmc*>fml).do_multilevel = self.method == "mcmc"
            if validation != NULL:
                fml.validation = validation
                (<fm_learn_mcmc*>fml).num_eval_cases = validation.num_cases
        
        fml.fm = &fm
        fml.max_target = self.train_set.max_rating
        fml.min_target = self.train_set.min_rating
        fml.meta = meta
        fml.task = 0  # regression

        fml.init()

        # regularization
        fm.reg0 = self.reg0
        fm.regw = self.reg1
        fm.regv = self.reg2
        if self.method in ["als", "mcmc"]:
            (<fm_learn_mcmc*>fml).w_lambda.init(fm.regw)
            (<fm_learn_mcmc*>fml).v_lambda.init(fm.regv)

        # learning rate
        (<fm_learn_sgd*>fml).learn_rate = self.learning_rate
        (<fm_learn_sgd*>fml).learn_rates.init(self.learning_rate)

        if self.verbose:
            fm.debug()
            fml.debug()

        # () learn
        fml.learn(dereference(train), dereference(validation))

        # store learned weights for prediction
        if self.k0:
            self.w0 = fm.w0
        if self.k1:
            for i in range(num_feature):
                w[i] = fm.w.get(i)
        if self.k2:
            for i in range(self.k2):
                for j in range(num_feature):
                    v[i, j] = <double>fm.v.get(i, j)


    def _fm_predict(self, user_idx, item_idx):
        uid = user_idx
        iid = item_idx + self.train_set.total_users
        score = 0.0
        if self.k0:
            score += self.w0
        if self.k1:
            score += self.w[uid] + self.w[iid]
        if self.k2:
            sum_ = self.v[:, uid] + self.v[:, iid]
            sum_sqr_ = self.v[:, uid] ** 2 + self.v[:, iid] ** 2
            score += 0.5 * (sum_ ** 2 - sum_sqr_).sum()
        return score


    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for that to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if item_idx is None:
            known_item_scores = np.fromiter(
                (self._fm_predict(user_idx, i) for i in range(self.train_set.total_items)), 
                dtype=np.double, 
                count=self.train_set.total_items
            )
            return known_item_scores
        else:
            return self._fm_predict(user_idx, item_idx)


# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example for Collaborative Topic Modeling"""

import cornac
from cornac.data import Reader
from cornac.datasets import citeulike
from cornac.eval_methods import RatioSplit
from cornac.data import TextModality
from cornac.data.text import BaseTokenizer


# CTR composes the LDA topic model with matrix factorization to model item (article) texts and user-item preferences
# The necessary data can be loaded as follows
docs, item_ids = citeulike.load_text()
feedback = citeulike.load_feedback(reader=Reader(item_set=item_ids))

# Instantiate a TextModality, it makes it convenient to work with text auxiliary information
# For more details, please refer to the tutorial on how to work with auxiliary data
item_text_modality = TextModality(
    corpus=docs,
    ids=item_ids,
    tokenizer=BaseTokenizer(sep=" ", stop_words="english"),
    max_vocab=8000,
    max_doc_freq=0.5,
)

# Define an evaluation method to split feedback into train and test sets
ratio_split = RatioSplit(
    data=feedback,
    test_size=0.2,
    exclude_unknowns=True,
    item_text=item_text_modality,
    verbose=True,
    seed=123,
    rating_threshold=0.5,
)

# Instantiate CTR model
ctr = cornac.models.CTR(k=50, max_iter=50, lambda_v=1)

# Use Recall@300 for evaluation
rec_300 = cornac.metrics.Recall(k=300)

# Put everything together into an experiment and run it
cornac.Experiment(eval_method=ratio_split, models=[ctr], metrics=[rec_300]).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example for Matrix Factorization with biases"""

import cornac
from cornac.datasets import movielens
from cornac.eval_methods import RatioSplit


# Load MovieLens 1M ratings
ml_1m = movielens.load_feedback(variant="1M")

# Define an evaluation method to split feedback into train and test sets
ratio_split = RatioSplit(
    data=ml_1m, test_size=0.2, exclude_unknowns=False, verbose=True
)

# Instantiate the global average baseline and MF model
global_avg = cornac.models.GlobalAvg()
mf = cornac.models.MF(
    k=10,
    max_iter=25,
    learning_rate=0.01,
    lambda_reg=0.02,
    use_bias=True,
    early_stop=True,
    verbose=True,
)

# Instantiate MAE and RMSE for evaluation
mae = cornac.metrics.MAE()
rmse = cornac.metrics.RMSE()

# Put everything together into an experiment and run it
cornac.Experiment(
    eval_method=ratio_split,
    models=[global_avg, mf],
    metrics=[mae, rmse],
    user_based=True,
).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import os
import numpy as np
from tqdm.auto import trange

from ..recommender import Recommender
from ...exception import ScoreException


os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

class NARRE(Recommender):
    """Neural Attentional Rating Regression with Review-level Explanations

    Parameters
    ----------
    name: string, default: 'NARRE'
        The name of the recommender model.

    embedding_size: int, default: 100
        Word embedding size

    id_embedding_size: int, default: 32
        User/item review id embedding size

    n_factors: int, default: 32
        The dimension of the user/item's latent factors.

    attention_size: int, default: 16
        Attention size

    kernel_sizes: list, default: [3]
        List of kernel sizes of conv2d

    n_filters: int, default: 64
        Number of filters

    dropout_rate: float, default: 0.5
        Dropout rate of neural network dense layers

    max_text_length: int, default: 50
        Maximum number of tokens in a review instance

    max_num_review: int, default: None
        Maximum number of reviews that you want to feed into training. By default, the model will be trained with all reviews.

    batch_size: int, default: 64
        Batch size

    max_iter: int, default: 10
        Max number of training epochs

    optimizer: string, optional, default: 'adam'
        Optimizer for training is either 'adam' or 'rmsprop'.

    learning_rate: float, optional, default: 0.001
        Initial value of learning rate for the optimizer.

    trainable: boolean, optional, default: True
        When False, the model will not be re-trained, and input of pre-trained parameters are required.

    verbose: boolean, optional, default: True
        When True, running logs are displayed.

    init_params: dictionary, optional, default: None
        Initial parameters, pretrained_word_embeddings could be initialized here, e.g., init_params={'pretrained_word_embeddings': pretrained_word_embeddings}

    seed: int, optional, default: None
        Random seed for weight initialization.
        If specified, training will take longer because of single-thread (no parallelization).

    References
    ----------
    * Chen, C., Zhang, M., Liu, Y., & Ma, S. (2018, April). Neural attentional rating regression with review-level explanations. In Proceedings of the 2018 World Wide Web Conference (pp. 1583-1592).
    """

    def __init__(
        self,
        name="NARRE",
        embedding_size=100,
        id_embedding_size=32,
        n_factors=32,
        attention_size=16,
        kernel_sizes=[3],
        n_filters=64,
        dropout_rate=0.5,
        max_text_length=50,
        max_num_review=None,
        batch_size=64,
        max_iter=10,
        optimizer='adam',
        learning_rate=0.001,
        trainable=True,
        verbose=True,
        init_params=None,
        seed=None,
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)
        self.seed = seed
        self.embedding_size = embedding_size
        self.id_embedding_size = id_embedding_size
        self.n_factors = n_factors
        self.attention_size = attention_size
        self.n_filters = n_filters
        self.kernel_sizes = kernel_sizes
        self.dropout_rate = dropout_rate
        self.max_text_length = max_text_length
        self.max_num_review = max_num_review
        self.batch_size = batch_size
        self.max_iter = max_iter
        self.optimizer = optimizer
        self.learning_rate = learning_rate
        # Init params if provided
        self.init_params = {} if init_params is None else init_params

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        if self.trainable:
            if not hasattr(self, "model"):
                from .narre import Model
                self.model = Model(
                    self.train_set.num_users,
                    self.train_set.num_items,
                    self.train_set.review_text.vocab,
                    self.train_set.global_mean,
                    n_factors=self.n_factors,
                    embedding_size=self.embedding_size,
                    id_embedding_size=self.id_embedding_size,
                    attention_size=self.attention_size,
                    kernel_sizes=self.kernel_sizes,
                    n_filters=self.n_filters,
                    dropout_rate=self.dropout_rate,
                    max_text_length=self.max_text_length,
                    pretrained_word_embeddings=self.init_params.get('pretrained_word_embeddings'),
                    verbose=self.verbose,
                    seed=self.seed,
                )
            self._fit_narre()

        return self

    def _fit_narre(self):
        import tensorflow as tf
        from tensorflow import keras
        from .narre import get_data
        loss = keras.losses.MeanSquaredError()
        if self.optimizer == 'rmsprop':
            optimizer = keras.optimizers.RMSprop(learning_rate=self.learning_rate)
        else:
            optimizer = keras.optimizers.Adam(learning_rate=self.learning_rate)
        train_loss = keras.metrics.Mean(name="loss")
        loop = trange(self.max_iter, disable=not self.verbose)
        for _ in loop:
            train_loss.reset_states()
            for i, (batch_users, batch_items, batch_ratings) in enumerate(self.train_set.uir_iter(self.batch_size, shuffle=True)):
                user_reviews, user_iid_reviews, user_num_reviews = get_data(batch_users, self.train_set, self.max_text_length, by='user', max_num_review=self.max_num_review)
                item_reviews, item_uid_reviews, item_num_reviews = get_data(batch_items, self.train_set, self.max_text_length, by='item', max_num_review=self.max_num_review)
                with tf.GradientTape() as tape:
                    predictions = self.model.graph(
                        [batch_users, batch_items, user_reviews, user_iid_reviews, user_num_reviews, item_reviews, item_uid_reviews, item_num_reviews],
                        training=True,
                    )
                    _loss = loss(batch_ratings, predictions)
                gradients = tape.gradient(_loss, self.model.graph.trainable_variables)
                optimizer.apply_gradients(zip(gradients, self.model.graph.trainable_variables))
                train_loss(_loss)
                if i % 10 == 0:
                    loop.set_postfix(loss=train_loss.result().numpy())
        loop.close()

        # save weights for predictions
        self.X, self.Y, self.W1, self.user_embedding, self.item_embedding, self.bu, self.bi, self.mu = self.model.get_weights(self.train_set, self.batch_size, max_num_review=self.max_num_review)
        if self.verbose:
            print("Learning completed!")


    def save(self, save_dir=None):
        """Save a recommender model to the filesystem.

        Parameters
        ----------
        save_dir: str, default: None
            Path to a directory for the model to be stored.

        """
        if save_dir is None:
            return
        model = self.model
        del self.model

        model_file = Recommender.save(self, save_dir)

        self.model = model
        self.model.save(model_file.replace(".pkl", ".cpt"))

        return model_file

    @staticmethod
    def load(model_path, trainable=False):
        """Load a recommender model from the filesystem.

        Parameters
        ----------
        model_path: str, required
            Path to a file or directory where the model is stored. If a directory is
            provided, the latest model will be loaded.

        trainable: boolean, optional, default: False
            Set it to True if you would like to finetune the model. By default, 
            the model parameters are assumed to be fixed after being loaded.
        
        Returns
        -------
        self : object
        """
        from tensorflow import keras
        model = Recommender.load(model_path, trainable)
        model.model = keras.models.load_model(model.load_from.replace(".pkl", ".cpt"))

        return model

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for that to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items
        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )
            h0 = (self.user_embedding[user_idx] + self.X[user_idx]) * (self.item_embedding + self.Y)
            known_item_scores = h0.dot(self.W1) + self.bu[user_idx] + self.bi + self.mu
            return known_item_scores.ravel()
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )
            h0 = (self.user_embedding[user_idx] + self.X[user_idx]) * (self.item_embedding[item_idx] + self.Y[item_idx])
            known_item_score = h0.dot(self.W1) + self.bu[user_idx] + self.bi[item_idx] + self.mu
            return known_item_score



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Fit to and evaluate CVAECF on the FilmTrust dataset"""

from cornac.data import GraphModality
from cornac.eval_methods import RatioSplit
from cornac.experiment import Experiment
from cornac import metrics
from cornac.models import CVAECF
from cornac.datasets import filmtrust

# In addition to learning from preference data, CVAECF further leverages users' auxiliary data (social network in this example).
# The necessary data can be loaded as follows
ratings = filmtrust.load_feedback()
trust = filmtrust.load_trust()

# Instantiate a GraphModality, it makes it convenient to work with graph (network) auxiliary information
# For more details, please refer to the tutorial on how to work with auxiliary data
user_graph_modality = GraphModality(data=trust)

# Define an evaluation method to split feedback into train and test sets
ratio_split = RatioSplit(
    data=ratings,
    test_size=0.2,
    rating_threshold=2.5,
    exclude_unknowns=True,
    verbose=True,
    user_graph=user_graph_modality,
    seed=123,
)

# Instantiate CVAECF model
cvaecf = CVAECF(
    z_dim=20,
    h_dim=20,
    autoencoder_structure=[40],
    learning_rate=0.001,
    n_epochs = 70,
    batch_size = 128,
    verbose=True,
    seed = 123
)


# Evaluation metrics
ndcg = metrics.NDCG(k=50)
rec = metrics.Recall(k=50)
pre = metrics.Precision(k=50)

# Put everything together into an experiment and run it
Experiment(
    eval_method=ratio_split,
    models=[cvaecf],
    metrics=[ndcg, pre, rec]
).run()

"""
Output:
       | NDCG@50 | Precision@50 | Recall@50 | Train (s) | Test (s)
------ + ------- + ------------ + --------- + --------- + --------
CVAECF |  0.4171 |       0.0781 |    0.8689 |   13.0752 |   1.4574
"""



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example for Variational Autoencoder for Collaborative Filtering, using the CiteULike dataset"""

import cornac
from cornac.datasets import citeulike
from cornac.eval_methods import RatioSplit


# Load user-item feedback
data = citeulike.load_feedback()

# Instantiate an evaluation method to split data into train and test sets.
ratio_split = RatioSplit(
    data=data,
    test_size=0.2,
    exclude_unknowns=True,
    verbose=True,
    seed=123,
    rating_threshold=0.5,
)

# Instantiate the VAECF model
vaecf = cornac.models.VAECF(
    k=10,
    autoencoder_structure=[20],
    act_fn="tanh",
    likelihood="mult",
    n_epochs=100,
    batch_size=100,
    learning_rate=0.001,
    beta=1.0,
    seed=123,
    use_gpu=True,
    verbose=True,
)

# Instantiate evaluation measures
rec_20 = cornac.metrics.Recall(k=20)
ndcg_20 = cornac.metrics.NDCG(k=20)
auc = cornac.metrics.AUC()

# Put everything together into an experiment and run it
cornac.Experiment(
    eval_method=ratio_split,
    models=[vaecf],
    metrics=[rec_20, ndcg_20, auc],
    user_based=True,
).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from libcpp.vector cimport vector

import numpy as np
from ...utils import get_rng
from ...utils.init_utils import gamma

ctypedef vector[vector[double]] Mat
ctypedef vector[double] dVec
ctypedef vector[int] iVec

cdef extern from "cpp_hpf.h":
    void pf_cpp(Mat & X, int & g, Mat & G_s, Mat & G_r, Mat & L_s, Mat & L_r, dVec & K_r, dVec & T_r, int maxiter)
    void hpf_cpp(Mat & X, int & g, Mat & G_s, Mat & G_r, Mat & L_s, Mat & L_r, dVec & K_r, dVec & T_r, int maxiter)




cpdef pf(Mat & X, int n_X, int d_X, int & k, int & iter_max, seed=None, init_param=None):
    n = n_X
    d = d_X

    #Hyper parameter setting
    a = 0.3
    a_ = 0.3
    a1 = 2.
    c = 0.3
    c_ = 0.3
    c1 = 1.
    b_ = 1
    d_ = 1
    k_s = a + k * a
    t_s = a + k * c
    
    rng = get_rng(seed)

    #Declare C++ variables
    cdef:
        Mat G_s, G_r, L_s, L_r

        #Parameters Initialization
    if init_param['G_s'] is None:
        G_s = gamma(a_, scale=b_ / a_, size=n * k, random_state=rng).reshape(n, k)
    else:
        G_s = init_param['G_s']

    # rate gamma_uk matrix
    if init_param['G_r'] is None:
        G_r = gamma(a_, scale=b_ / a_, size=n * k, random_state=rng).reshape(n, k)
    else:
        G_r = init_param['G_r']

    ## shape lamda_ik matrix
    if init_param['L_s'] is None:
        L_s = gamma(c_, scale=d_ / c_, size=d * k, random_state=rng).reshape(d, k)
    else:
        L_s = init_param['L_s']

    #rate lamda_ik matrix
    if init_param['L_r'] is None:
        L_r = gamma(c_, scale=d_ / c_, size=d * k, random_state=rng).reshape(d, k)
    else:
        L_r = init_param['L_r']

    K_r = np.repeat(1.0, n)
    T_r = np.repeat(1.0, d)

    print('Learning...')
    pf_cpp(X, k, G_s, G_r, L_s, L_r, K_r, T_r, iter_max)
    print('Learning completed!')

    res = {
        "Z": np.array(G_s) / np.array(G_r), 
        "W": np.array(L_s) / np.array(L_r),
        "G_s": np.array(G_s),
        "G_r": np.array(G_r),
        "L_s": np.array(L_s),
        "L_r": np.array(L_r),
    }

    return res

#Hierarchical Poisson factorization
cpdef hpf(Mat & X, int n_X, int d_X, int & k, int & iter_max, seed=None, init_param = None):
    n = n_X
    d = d_X

    #Hyper parameter setting
    a = 0.3
    #a_ = 0.3
    a_ = 100.
    a1 = 2.
    c = 0.3
    #c_ = 0.3
    c_ = 100.
    c1 = 1.
    #b_ = 1
    b_ = 0.3
    #d_ = 1
    d_ = 0.3
    
    rng = get_rng(seed)

    #Declare C++ variables
    cdef:
        Mat G_s, G_r, L_s, L_r

        #Parameters Initialization
    if init_param['G_s'] is None:
        G_s = gamma(a_, scale=b_ / a_, size=n * k, random_state=rng).reshape(n, k)
    else:
        G_s = init_param['G_s']

    # rate gamma_uk matrix
    if init_param['G_r'] is None:
        G_r = gamma(a_, scale=b_ / a_, size=n * k, random_state=rng).reshape(n, k)
    else:
        G_r = init_param['G_r']

    ## shape lamda_ik matrix
    if init_param['L_s'] is None:
        L_s = gamma(c_, scale=d_ / c_, size=d * k, random_state=rng).reshape(d, k)
    else:
        L_s = init_param['L_s']

    #rate lamda_ik matrix
    if init_param['L_r'] is None:
        L_r = gamma(c_, scale=d_ / c_, size=d * k, random_state=rng).reshape(d, k)
    else:
        L_r = init_param['L_r']

    K_r = np.repeat(1.0, n)
    T_r = np.repeat(1.0, d)

    print('Learning...')
    hpf_cpp(X, k, G_s, G_r, L_s, L_r, K_r, T_r, iter_max)
    print('Learning completed!')

    res = {
        "Z": np.array(G_s) / np.array(G_r), 
        "W": np.array(L_s) / np.array(L_r),
        "G_s": np.array(G_s),
        "G_r": np.array(G_r),
        "L_s": np.array(L_s),
        "L_r": np.array(L_r),
    }

    return res



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example for Social Bayesian Personalized Ranking (SBPR) with Epinions dataset"""

import cornac
from cornac.data import Reader, GraphModality
from cornac.datasets import epinions
from cornac.eval_methods import RatioSplit


# SBPR integrates user social network into Bayesian Personalized Ranking.
# The necessary data can be loaded as follows
feedback = epinions.load_feedback(
    Reader(bin_threshold=4.0)
)  # feedback is binarised (turned into implicit) using Reader.
trust = epinions.load_trust()

# Instantiate a GraphModality, it makes it convenient to work with graph (network) auxiliary information
# For more details, please refer to the tutorial on how to work with auxiliary data
user_graph_modality = GraphModality(data=trust)

# Define an evaluation method to split feedback into train and test sets
ratio_split = RatioSplit(
    data=feedback,
    test_size=0.1,
    rating_threshold=0.5,
    exclude_unknowns=True,
    verbose=True,
    user_graph=user_graph_modality,
)

# Instantiate SBPR model
sbpr = cornac.models.SBPR(
    k=10,
    max_iter=50,
    learning_rate=0.001,
    lambda_u=0.015,
    lambda_v=0.025,
    lambda_b=0.01,
    verbose=True,
)

# Use Recall@10 for evaluation
rec_10 = cornac.metrics.Recall(k=10)

# Put everything together into an experiment and run it
cornac.Experiment(eval_method=ratio_split, models=[sbpr], metrics=[rec_10]).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

# cython: language_level=3

cimport cython
from cython cimport floating, integral
from cython.parallel import parallel, prange
from libc.math cimport exp, floor

import numpy as np
cimport numpy as np
from scipy.sparse import csr_matrix
from tqdm.auto import trange

from ..recommender import Recommender
from ..bpr.recom_bpr import BPR
from ..bpr.recom_bpr cimport RNGVector, has_non_zero


cdef extern from "../bpr/recom_bpr.h" namespace "recom_bpr" nogil:
    cdef int get_thread_num()


class SBPR(BPR):
    """Social Bayesian Personalized Ranking.

    Parameters
    ----------
    k: int, optional, default: 10
        The dimension of the latent factors.

    max_iter: int, optional, default: 100
        Maximum number of iterations or the number of epochs for SGD.

    learning_rate: float, optional, default: 0.001
        The learning rate for SGD.

    lambda_reg: float, optional, default: 0.001
        The regularization hyper-parameter.

    num_threads: int, optional, default: 0
        Number of parallel threads for training. If num_threads=0, all CPU cores will be utilized.
        If seed is not None, num_threads=1 to remove randomness from parallelization.

    trainable: boolean, optional, default: True
        When False, the model will not be re-trained, and input of pre-trained parameters are required.

    verbose: boolean, optional, default: True
        When True, some running logs are displayed.

    init_params: dictionary, optional, default: None
        Initial parameters, e.g., init_params = {'U': user_factors, 'V': item_factors, 'Bi': item_biases}

    seed: int, optional, default: None
        Random seed for weight initialization.
        If specified, training will take longer because of single-thread (no parallelization).

    References
    ----------
    * Zhao, T., McAuley, J., & King, I. (2014, November). Leveraging social connections to improve \
    personalized ranking for collaborative filtering. CIKM 2014 (pp. 261-270).
    """

    def __init__(
        self, 
        name='SBPR', 
        k=10,
        max_iter=100, 
        learning_rate=0.001,
        lambda_u=0.01, 
        lambda_v=0.01, 
        lambda_b=0.01,
        num_threads=0, 
        trainable=True, 
        verbose=False, 
        init_params=None, 
        seed=None
    ):
        super().__init__(
            name=name, 
            k=k, 
            max_iter=max_iter, 
            learning_rate=learning_rate, 
            num_threads=num_threads, 
            trainable=trainable, 
            verbose=verbose, 
            init_params=init_params, 
            seed=seed
        )
        self.lambda_u = lambda_u
        self.lambda_v = lambda_v
        self.lambda_b = lambda_b

    def _prepare_social_data(self):
        X = self.train_set.matrix # csr_matrix
        n_users, n_items = self.train_set.num_users, self.train_set.num_items

        # construct social feedback in the sparse format
        (rid, cid, val) = self.train_set.user_graph.get_train_triplet(
            self.train_set.user_indices, self.train_set.user_indices
        )
        Y = csr_matrix((val, (rid, cid)), shape=(n_users, n_users))
        social_item_ids = []
        social_item_counts = []
        social_indptr = [0]
        for uid in trange(n_users, disable=not self.verbose, desc='Building social data'):
            real_pos_items = np.unique(X[uid].indices)
            social_pos_items, counts = np.unique(X[Y[uid].indices].indices, return_counts=True)
            mask = np.in1d(social_pos_items, real_pos_items, assume_unique=True)
            social_item_ids.extend(social_pos_items[~mask])
            social_item_counts.extend(counts[~mask])
            social_indptr.append(len(social_item_ids))

        social_item_ids = np.asarray(social_item_ids).astype(X.indices.dtype)
        social_item_counts = np.asarray(social_item_counts).astype(X.indices.dtype)
        social_indptr = np.asarray(social_indptr).astype(X.indices.dtype)

        return social_item_ids, social_item_counts, social_indptr

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        self._init()

        if not self.trainable:
            return self
        
        X, user_counts, user_ids = self._prepare_data()
        s_item_ids, s_item_counts, s_indptr = self._prepare_social_data()

        # construct random generators
        cdef:
            int num_threads = self.num_threads
            RNGVector rng_pos = RNGVector(num_threads, len(user_ids) - 1, self.rng.randint(2 ** 31))
            RNGVector rng_neg = RNGVector(num_threads, train_set.num_items - 1, self.rng.randint(2 ** 31))

        # start training
        with trange(self.max_iter, disable=not self.verbose) as progress:
            for epoch in progress:
                skipped = self._fit_sgd(rng_pos, rng_neg, num_threads,
                                        user_ids, X.indices, X.indptr,
                                        s_item_ids, s_item_counts, s_indptr,
                                        self.u_factors, self.i_factors, self.i_biases)
                progress.set_postfix({"skipped": "%.2f%%" % (100.0 * skipped / len(user_ids))})
        if self.verbose:
            print('Optimization finished!')

        return self

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _fit_sgd(self, RNGVector rng_pos, RNGVector rng_neg, int num_threads,
                 integral[:] user_ids, integral[:] item_ids, integral[:] indptr,
                 integral[:] social_item_ids, integral[:] social_item_counts, integral[:] social_indptr,
                 floating[:, :] U, floating[:, :] V, floating[:] B):
        """Fit the model parameters (U, V, B) with SGD
        """
        cdef:
            long num_samples = len(user_ids)
            long num_items = self.train_set.num_items
            long s, i_index, k_index, skipped = 0
            int f, u_id, i_id, j_id, k_id, n_social_items, thread_id
            floating u_temp, k_rand
            floating z, score # for BPR formula
            floating z_ik, z_kj, score_ik, score_kj, s_uk # for SBPR-2 formula

            floating lr = self.learning_rate
            floating lbd_u = self.lambda_u
            floating lbd_v = self.lambda_v
            floating lbd_b = self.lambda_b
            int factors = self.k

            floating * user
            floating * item_i
            floating * item_j
            floating * item_k

        with nogil, parallel(num_threads=num_threads):
            thread_id = get_thread_num()

            for s in prange(num_samples, schedule='guided'):
                i_index = rng_pos.generate(thread_id)
                u_id = user_ids[i_index]
                i_id = item_ids[i_index]
                j_id = rng_neg.generate(thread_id)

                # sample social item k_id for given user u_id
                n_social_items = social_indptr[u_id + 1] - social_indptr[u_id]
                k_rand = <float>rng_neg.generate(thread_id) / num_items # uniform between [0.0, 1.0)
                k_index = social_indptr[u_id] + <int>floor(k_rand * n_social_items)
                k_id = social_item_ids[k_index]

                # if the user has liked the item j,
                # else if item j is also a social item,
                # skip this for now
                if has_non_zero(indptr, item_ids, u_id, j_id) or (j_id == k_id):
                    skipped += 1
                    continue

                # get pointers to the relevant factors
                user = &U[u_id, 0]
                item_i, item_j, item_k = &V[i_id, 0], &V[j_id, 0], &V[k_id, 0]

                # if no social item for given user uid, update factors based on BPR formula
                if n_social_items == 0:
                    # compute the score
                    score = B[i_id] - B[j_id]
                    for f in range(factors):
                        score = score + user[f] * (item_i[f] - item_j[f])
                    z = 1.0 / (1.0 + exp(score))

                    # update the factors via sgd.
                    for f in range(factors):
                        u_temp = user[f]
                        user[f] += lr * (z * (item_i[f] - item_j[f]) - lbd_u* user[f])
                        item_i[f] += lr * (z * u_temp - lbd_v * item_i[f])
                        item_j[f] += lr * (-z * u_temp - lbd_v * item_j[f])

                    # update item biases
                    B[i_id] += lr * (z - lbd_b * B[i_id])
                    B[j_id] += lr * (-z - lbd_b * B[j_id])

                    continue

                # found social feedback, update factors based on SBPR-2 formula
                # compute the scores
                score_ik = B[i_id] - B[k_id]
                score_kj = B[k_id] - B[j_id]
                for f in range(factors):
                    score_ik = score_ik + user[f] * (item_i[f] - item_k[f])
                    score_kj = score_kj + user[f] * (item_k[f] - item_j[f])
                s_uk = 1.0 / (1.0 + social_item_counts[k_index])
                z_ik = 1.0 / (1.0 + exp(score_ik * s_uk))
                z_kj = 1.0 / (1.0 + exp(score_kj))

                # update the factors via sgd.
                for f in range(factors):
                    u_temp = user[f]
                    user[f] += lr * (z_ik * (item_i[f] - item_k[f]) * s_uk + 
                                     z_kj * (item_k[f] - item_j[f]) -
                                     lbd_u * user[f])
                    item_i[f] += lr * (z_ik * u_temp * s_uk - lbd_v * item_i[f])
                    item_j[f] += lr * (-z_kj * u_temp - lbd_v * item_j[f])
                    item_k[f] += lr * (z_kj * u_temp - z_ik * u_temp * s_uk - lbd_v * item_k[f])

                # update item biases
                B[i_id] += lr * (z_ik * s_uk - lbd_b * B[i_id])
                B[j_id] += lr * (-z_kj - lbd_b * B[j_id])
                B[k_id] += lr * (z_kj - z_ik * s_uk - lbd_b * B[k_id])

        return skipped



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================


# cython: language_level=3, boundscheck=False, wraparound=False, cdivision=True

cimport cython
from cython cimport floating
from libc.math cimport fabs, sqrt

cimport numpy as np
import numpy as np

ctypedef fused integral:
    int
    long long

def inplace_csr_row_normalize_l1(X):
    """Inplace row normalize using the l1 norm"""
    _inplace_csr_row_normalize_l1(X.data, X.shape, X.indices, X.indptr)

def _inplace_csr_row_normalize_l1(np.ndarray[floating, ndim=1] X_data,
                                  shape,
                                  np.ndarray[integral, ndim=1] X_indices,
                                  np.ndarray[integral, ndim=1] X_indptr):
    cdef unsigned long long n_samples = shape[0]
    cdef unsigned long long n_features = shape[1]

    # the column indices for row i are stored in:
    #    indices[indptr[i]:indices[i+1]]
    # and their corresponding values are stored in:
    #    data[indptr[i]:indptr[i+1]]
    cdef np.npy_intp i, j
    cdef double sum_

    for i in range(n_samples):
        sum_ = 0.0

        for j in range(X_indptr[i], X_indptr[i + 1]):
            sum_ += fabs(X_data[j])

        if sum_ == 0.0:
            # do not normalize empty rows (can happen if CSR is not pruned
            # correctly)
            continue

        for j in range(X_indptr[i], X_indptr[i + 1]):
            X_data[j] /= sum_

def inplace_csr_row_normalize_l2(X):
    """Inplace row normalize using the l2 norm"""
    _inplace_csr_row_normalize_l2(X.data, X.shape, X.indices, X.indptr)

def _inplace_csr_row_normalize_l2(np.ndarray[floating, ndim=1] X_data,
                                  shape,
                                  np.ndarray[integral, ndim=1] X_indices,
                                  np.ndarray[integral, ndim=1] X_indptr):
    cdef integral n_samples = shape[0]
    cdef integral n_features = shape[1]

    cdef np.npy_intp i, j
    cdef double sum_

    for i in range(n_samples):
        sum_ = 0.0

        for j in range(X_indptr[i], X_indptr[i + 1]):
            sum_ += (X_data[j] * X_data[j])

        if sum_ == 0.0:
            # do not normalize empty rows (can happen if CSR is not pruned
            # correctly)
            continue

        sum_ = sqrt(sum_)

        for j in range(X_indptr[i], X_indptr[i + 1]):
            X_data[j] /= sum_



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example on how to train and evaluate a model with provided train and test sets"""

from cornac.data import Reader
from cornac.eval_methods import BaseMethod
from cornac.models import MF
from cornac.metrics import MAE, RMSE
from cornac.utils import cache


# Download MovieLens 100K provided train and test sets
reader = Reader()
train_data = reader.read(
    cache(url="http://files.grouplens.org/datasets/movielens/ml-100k/u1.base")
)
test_data = reader.read(
    cache(url="http://files.grouplens.org/datasets/movielens/ml-100k/u1.test")
)

# Instantiate a Base evaluation method using the provided train and test sets
eval_method = BaseMethod.from_splits(
    train_data=train_data, test_data=test_data, exclude_unknowns=False, verbose=True
)

# Instantiate the MF model
mf = MF(
    k=10,
    max_iter=25,
    learning_rate=0.01,
    lambda_reg=0.02,
    use_bias=True,
    early_stop=True,
    verbose=True,
)

# Evaluation
test_result, val_result = eval_method.evaluate(
    model=mf, metrics=[MAE(), RMSE()], user_based=True
)
print(test_result)



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

# cython: language_level=3

from cython cimport floating, integral
from libcpp.vector cimport vector


cdef int get_key(int i_id, int j_id) nogil

cdef floating get_score(floating[:, :, :] G, int dim1, int dim2, int dim3,
                        floating[:, :] U, floating[:, :] I, floating[:, :] A,
                        int u_idx, int i_idx, int a_idx) nogil



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example for Multi-Task Explainable Recommendation"""

from cornac.datasets import amazon_toy
from cornac.data import SentimentModality
from cornac.eval_methods import RatioSplit
from cornac.metrics import NDCG, RMSE
from cornac.models import MTER
from cornac import Experiment


# Load rating and sentiment information
data = amazon_toy.load_feedback()
sentiment = amazon_toy.load_sentiment()

# Instantiate a SentimentModality, it makes it convenient to work with sentiment information
md = SentimentModality(data=sentiment)

# Define an evaluation method to split feedback into train and test sets
eval_method = RatioSplit(
    data,
    test_size=0.2,
    rating_threshold=1.0,
    sentiment=md,
    exclude_unknowns=True,
    verbose=True,
    seed=123,
)

# Instantiate the MTER model
mter = MTER(
    n_user_factors=15,
    n_item_factors=15,
    n_aspect_factors=12,
    n_opinion_factors=12,
    n_bpr_samples=1000,
    n_element_samples=50,
    lambda_reg=0.1,
    lambda_bpr=10,
    max_iter=100000,
    lr=0.1,
    verbose=True,
    seed=123,
)

# Instantiate and run an experiment
Experiment(
    eval_method=eval_method,
    models=[mter],
    metrics=[RMSE(), NDCG(k=10), NDCG(k=20), NDCG(k=50), NDCG(k=100)],
).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example to run Probabilistic Matrix Factorization (PMF) model with Ratio Split evaluation strategy"""

import cornac
from cornac.datasets import movielens
from cornac.eval_methods import RatioSplit
from cornac.models import PMF


# Load the MovieLens 100K dataset
ml_100k = movielens.load_feedback()

# Instantiate an evaluation method.
ratio_split = RatioSplit(
    data=ml_100k, test_size=0.2, rating_threshold=4.0, exclude_unknowns=False
)

# Instantiate a PMF recommender model.
pmf = PMF(k=10, max_iter=100, learning_rate=0.001, lambda_reg=0.001)

# Instantiate evaluation metrics.
mae = cornac.metrics.MAE()
rmse = cornac.metrics.RMSE()
rec_20 = cornac.metrics.Recall(k=20)
pre_20 = cornac.metrics.Precision(k=20)

# Instantiate and then run an experiment.
cornac.Experiment(
    eval_method=ratio_split,
    models=[pmf],
    metrics=[mae, rmse, rec_20, pre_20],
    user_based=True,
).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example for Explicit Factor Models"""

import cornac
from cornac.datasets import amazon_toy
from cornac.data import SentimentModality
from cornac.eval_methods import RatioSplit


# Load rating and sentiment information
rating = amazon_toy.load_feedback()
sentiment = amazon_toy.load_sentiment()

# Instantiate a SentimentModality, it makes it convenient to work with sentiment information
md = SentimentModality(data=sentiment)

# Define an evaluation method to split feedback into train and test sets
split_data = RatioSplit(
    data=rating,
    test_size=0.15,
    exclude_unknowns=True,
    verbose=True,
    sentiment=md,
    seed=123,
)

# Instantiate the EFM model
efm = cornac.models.EFM(
    num_explicit_factors=40,
    num_latent_factors=60,
    num_most_cared_aspects=15,
    rating_scale=5.0,
    alpha=0.85,
    lambda_x=1,
    lambda_y=1,
    lambda_u=0.01,
    lambda_h=0.01,
    lambda_v=0.01,
    max_iter=100,
    num_threads=1,
    trainable=True,
    verbose=True,
    seed=123,
)

# Instantiate evaluation metrics
rmse = cornac.metrics.RMSE()
ndcg_50 = cornac.metrics.NDCG(k=50)
auc = cornac.metrics.AUC()

# Put everything together into an experiment and run it
cornac.Experiment(
    eval_method=split_data, models=[efm], metrics=[rmse, ndcg_50, auc]
).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np

NUM_FMT = "{:.4f}"


def _table_format(data, headers=None, index=None, extra_spaces=0, h_bars=None):
    if headers is not None:
        data.insert(0, headers)
    if index is not None:
        index.insert(0, "")
        for idx, row in zip(index, data):
            row.insert(0, idx)

    column_widths = np.asarray([[len(str(v)) for v in row] for row in data]).max(axis=0)

    row_fmt = (
        " | ".join(["{:>%d}" % (w + extra_spaces) for w in column_widths][1:]) + "\n"
    )
    if index is not None:
        row_fmt = "{:<%d} | " % (column_widths[0] + extra_spaces) + row_fmt

    output = ""
    for i, row in enumerate(data):
        if h_bars is not None and i in h_bars:
            output += row_fmt.format(
                *["-" * (w + extra_spaces) for w in column_widths]
            ).replace("|", "+")
        output += row_fmt.format(*row)
    return output


class Result:
    """
    Result Class for a single model

    Parameters
    ----------
    model_name: string, required
        The name of the recommender model.

    metric_avg_results: :obj:`OrderedDict`, required
        A dictionary containing the average result per-metric.

    metric_user_results: :obj:`OrderedDict`, required
        A dictionary containing the average result per-user across different metrics.
    """

    def __init__(self, model_name, metric_avg_results, metric_user_results):
        self.model_name = model_name
        self.metric_avg_results = metric_avg_results
        self.metric_user_results = metric_user_results

    def __str__(self):
        headers = list(self.metric_avg_results.keys())
        data = [[NUM_FMT.format(v) for v in self.metric_avg_results.values()]]
        return _table_format(data, headers, index=[self.model_name], h_bars=[1])


class CVResult(list):
    """
    Cross Validation Result Class for a single model. A list of :obj:`cornac.experiment.Result`.
    
    Parameters
    ----------
    model_name: string, required
        The name of the recommender model.
    """

    def __init__(self, model_name):
        super().__init__()
        self.model_name = model_name

    def __str__(self):
        return "[{}]\n{}".format(self.model_name, self.table)

    def organize(self):
        headers = list(self[0].metric_avg_results.keys())
        data, index = [], []
        for f, r in enumerate(self):
            data.append([r.metric_avg_results[m] for m in headers])
            index.append("Fold %d" % f)

        data = np.asarray(data)
        mean, std = data.mean(axis=0), data.std(axis=0)
        data = np.vstack([data, mean, std])
        data = [[NUM_FMT.format(v) for v in row] for row in data]
        index.extend(["Mean", "Std"])
        self.table = _table_format(data, headers, index, h_bars=[1, len(data) - 1])


class ExperimentResult(list):
    """
    Result Class for an Experiment. A list of :obj:`cornac.experiment.Result`. 
    """

    def __str__(self):
        headers = list(self[0].metric_avg_results.keys())
        data, index = [], []
        for r in self:
            data.append([NUM_FMT.format(r.metric_avg_results[m]) for m in headers])
            index.append(r.model_name)
        return _table_format(data, headers, index, h_bars=[1])


class CVExperimentResult(ExperimentResult):
    """
    Result Class for a cross-validation Experiment.
    """

    def __str__(self):
        return "\n".join([r.__str__() for r in self])



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Fit to and evaluate PCRL [1] on the Office Amazon dataset.
[1] Salah, Aghiles, and Hady W. Lauw. Probabilistic Collaborative Representation Learning\
    for Personalized Item Recommendation. In UAI 2018.
"""

from cornac.data import GraphModality
from cornac.eval_methods import RatioSplit
from cornac.experiment import Experiment
from cornac import metrics
from cornac.models import PCRL
from cornac.datasets import amazon_office as office


# PCRL jointly models user preferences and performs deep item features learning from auxiliary data (e.g., item relations).
# The necessary data can be loaded as follows
ratings = office.load_feedback()
contexts = office.load_graph()

# Instantiate a GraphModality, it makes it convenient to work with graph (network) auxiliary information
# For more details, please refer to the tutorial on how to work with auxiliary data
item_graph_modality = GraphModality(data=contexts)

# Define an evaluation method to split feedback into train and test sets
ratio_split = RatioSplit(
    data=ratings,
    test_size=0.2,
    rating_threshold=3.5,
    exclude_unknowns=True,
    verbose=True,
    item_graph=item_graph_modality,
)

# Instantiate PCRL
pcrl = PCRL(k=100, z_dims=[300], max_iter=300, learning_rate=0.001)

# Evaluation metrics
nDgc = metrics.NDCG(k=-1)
rec = metrics.Recall(k=20)
pre = metrics.Precision(k=20)

# Put everything together into an experiment and run it
Experiment(eval_method=ratio_split, models=[pcrl], metrics=[nDgc, rec, pre]).run()


"""
Output:
     | NDCG@-1 | Recall@20 | Precision@20 | Train (s) | Test (s)
---- + ------- + --------- + ------------ + --------- + --------
pcrl |  0.1922 |    0.0862 |       0.0148 | 2591.4878 |   4.0957

*Results may change slightly from one run to another due to different random initial parameters
"""



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example for Collaborative Filtering for Implicit Feedback Datasets (Citeulike)"""

import cornac
from cornac.data import Reader
from cornac.datasets import citeulike
from cornac.eval_methods import RatioSplit

# Load user-item feedback
_, item_ids = citeulike.load_text()
data = citeulike.load_feedback(reader=Reader(item_set=item_ids))

# Instantiate an evaluation method to split data into train and test sets.
ratio_split = RatioSplit(
    data=data,
    test_size=0.2,
    exclude_unknowns=True,
    verbose=True,
    seed=123,
    rating_threshold=0.5,
)

# Instantiate the WMF model
wmf = cornac.models.WMF(
    k=50,
    max_iter=50,
    learning_rate=0.001,
    lambda_u=0.01,
    lambda_v=0.01,
    verbose=True,
    seed=123,
)

# Use Recall@300 for evaluation
rec_300 = cornac.metrics.Recall(k=300)

# Instantiate and run an experiment
cornac.Experiment(
    eval_method=ratio_split, models=[wmf], metrics=[rec_300], user_based=True
).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Example for Factorization Machines with MovieLens 100K dataset"""


import cornac
from cornac.datasets import movielens
from cornac.eval_methods import RatioSplit
from cornac.models import FM


feedback = movielens.load_feedback(variant="100K")

ratio_split = RatioSplit(
    data=feedback,
    test_size=0.1,
    val_size=0.1,
    exclude_unknowns=True,
    verbose=True,
    seed=42,
)

models = [
    FM(k0=1, k1=1, k2=8, method="sgd", max_iter=100, seed=42, name="sgd"),
    FM(k0=1, k1=1, k2=8, method="sgda", max_iter=100, seed=42, name="sgda"),
    FM(k0=1, k1=1, k2=8, method="als", max_iter=100, seed=42, name="als"),
    FM(k0=1, k1=1, k2=8, method="mcmc", max_iter=100, seed=42, name="mcmc"),
]

cornac.Experiment(
    eval_method=ratio_split,
    models=models,
    metrics=[cornac.metrics.RMSE()],
    user_based=False,
).run()



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from scipy.stats import rankdata


class RankingMetric:
    """Ranking Metric.

    Attributes
    ----------
    type: string, value: 'ranking'
        Type of the metric, e.g., "ranking", "rating".

    name: string, default: None
        Name of the measure.

    k: int or list, optional, default: -1 (all)
        The number of items in the top@k list.
        If None, all items will be considered.

    """

    def __init__(self, name=None, k=-1, higher_better=True):
        assert hasattr(k, "__len__") or k == -1 or k > 0

        self.type = "ranking"
        self.name = name
        self.k = k
        self.higher_better = higher_better

    def compute(self, **kwargs):
        raise NotImplementedError()


class NDCG(RankingMetric):
    """Normalized Discount Cumulative Gain.

    Parameters
    ----------
    k: int or list, optional, default: -1 (all)
        The number of items in the top@k list.
        If None, all items will be considered.

    References
    ----------
    https://en.wikipedia.org/wiki/Discounted_cumulative_gain

    """

    def __init__(self, k=-1):
        RankingMetric.__init__(self, name="NDCG@{}".format(k), k=k)

    @staticmethod
    def dcg_score(gt_pos, pd_rank, k=-1):
        """Compute Discounted Cumulative Gain score.

        Parameters
        ----------
        gt_pos: Numpy array
            Binary vector of positive items.

        pd_rank: Numpy array
            Item ranking prediction.

        k: int, optional, default: -1 (all)
            The number of items in the top@k list.
            If None, all items will be considered.

        Returns
        -------
        dcg: A scalar
            Discounted Cumulative Gain score.

        """
        if k > 0:
            truncated_pd_rank = pd_rank[:k]
        else:
            truncated_pd_rank = pd_rank

        ranked_scores = np.take(gt_pos, truncated_pd_rank)
        gain = 2 ** ranked_scores - 1
        discounts = np.log2(np.arange(len(ranked_scores)) + 2)

        return np.sum(gain / discounts)

    def compute(self, gt_pos, pd_rank, **kwargs):
        """Compute Normalized Discounted Cumulative Gain score.

        Parameters
        ----------
        gt_pos: Numpy array
            Binary vector of positive items.

        pd_rank: Numpy array
            Item ranking prediction.

        **kwargs: For compatibility

        Returns
        -------
        ndcg: A scalar
            Normalized Discounted Cumulative Gain score.

        """
        dcg = self.dcg_score(gt_pos, pd_rank, self.k)
        idcg = self.dcg_score(gt_pos, np.argsort(gt_pos)[::-1], self.k)
        ndcg = dcg / idcg

        return ndcg


class NCRR(RankingMetric):
    """Normalized Cumulative Reciprocal Rank.

    Parameters
    ----------
    k: int or list, optional, default: -1 (all)
        The number of items in the top@k list.
        If None, all items will be considered.

    """

    def __init__(self, k=-1):
        RankingMetric.__init__(self, name="NCRR@{}".format(k), k=k)

    def compute(self, gt_pos, pd_rank, **kwargs):
        """Compute Normalized Cumulative Reciprocal Rank score.

        Parameters
        ----------
        gt_pos: Numpy array
            Binary vector of positive items.

        pd_rank: Numpy array
            Item ranking prediction.

        **kwargs: For compatibility

        Returns
        -------
        ncrr: A scalar
            Normalized Cumulative Reciprocal Rank score.

        """
        if self.k > 0:
            truncated_pd_rank = pd_rank[: self.k]
        else:
            truncated_pd_rank = pd_rank

        gt_pos_items = np.nonzero(gt_pos > 0)

        # Compute CRR
        rec_rank = np.where(np.in1d(truncated_pd_rank, gt_pos_items))[0]
        if len(rec_rank) == 0:
            return 0.0
        rec_rank = rec_rank + 1  # +1 because indices starts from 0 in python
        crr = np.sum(1.0 / rec_rank)

        # Compute Ideal CRR
        max_nb_pos = min(len(gt_pos_items[0]), len(truncated_pd_rank))
        ideal_rank = np.arange(max_nb_pos)
        ideal_rank = ideal_rank + 1  # +1 because indices starts from 0 in python
        icrr = np.sum(1.0 / ideal_rank)

        # Compute nDCG
        ncrr_i = crr / icrr

        return ncrr_i


class MRR(RankingMetric):
    """Mean Reciprocal Rank.

    References
    ----------
    https://en.wikipedia.org/wiki/Mean_reciprocal_rank
    """

    def __init__(self):
        RankingMetric.__init__(self, name="MRR")

    def compute(self, gt_pos, pd_rank, **kwargs):
        """Compute Mean Reciprocal Rank score.

        Parameters
        ----------
        gt_pos: Numpy array
            Binary vector of positive items.

        pd_rank: Numpy array
            Item ranking prediction.

        **kwargs: For compatibility

        Returns
        -------
        mrr: A scalar
            Mean Reciprocal Rank score.

        """
        gt_pos_items = np.nonzero(gt_pos > 0)
        matched_items = np.nonzero(np.in1d(pd_rank, gt_pos_items))[0]

        if len(matched_items) == 0:
            raise ValueError(
                "No matched between ground-truth items and recommendations"
            )

        mrr = np.divide(
            1, (matched_items[0] + 1)
        )  # +1 because indices start from 0 in python
        return mrr


class MeasureAtK(RankingMetric):
    """Measure at K.

    Attributes
    ----------
    k: int or list, optional, default: -1 (all)
        The number of items in the top@k list.
        If None, all items will be considered.

    """

    def __init__(self, name=None, k=-1):
        RankingMetric.__init__(self, name, k)

    def compute(self, gt_pos, pd_rank, **kwargs):
        """Compute TP, TP+FN, and TP+FP.

        Parameters
        ----------
        gt_pos: Numpy array
            Binary vector of positive items.

        pd_rank: Numpy array
            Item ranking prediction.

        **kwargs: For compatibility

        Returns
        -------
        tp: A scalar
            True positive.

        tp_fn: A scalar
            True positive + false negative.

        tp_fp: A scalar
            True positive + false positive.

        """
        if self.k > 0:
            truncated_pd_rank = pd_rank[: self.k]
        else:
            truncated_pd_rank = pd_rank

        pred = np.zeros_like(gt_pos)
        pred[truncated_pd_rank] = 1

        tp = np.sum(pred * gt_pos)
        tp_fn = np.sum(gt_pos)
        tp_fp = np.sum(pred)

        return tp, tp_fn, tp_fp


class Precision(MeasureAtK):
    """Precision@K.

    Parameters
    ----------
    k: int or list, optional, default: -1 (all)
        The number of items in the top@k list.
        If None, all items will be considered.

    """

    def __init__(self, k=-1):
        super().__init__(name="Precision@{}".format(k), k=k)

    def compute(self, gt_pos, pd_rank, **kwargs):
        """Compute Precision score.

        Parameters
        ----------
        gt_pos: Numpy array
            Binary vector of positive items.

        pd_rank: Numpy array
            Item ranking prediction.

        **kwargs: For compatibility

        Returns
        -------
        res: A scalar
            Precision score.

        """
        tp, _, tp_fp = MeasureAtK.compute(self, gt_pos, pd_rank, **kwargs)
        return tp / tp_fp


class Recall(MeasureAtK):
    """Recall@K.

    Parameters
    ----------
    k: int or list, optional, default: -1 (all)
        The number of items in the top@k list.
        If None, all items will be considered.

    """

    def __init__(self, k=-1):
        super().__init__(name="Recall@{}".format(k), k=k)

    def compute(self, gt_pos, pd_rank, **kwargs):
        """Compute Recall score.

        Parameters
        ----------
        gt_pos: Numpy array
            Binary vector of positive items.

        pd_rank: Numpy array
            Item ranking prediction.

        **kwargs: For compatibility

        Returns
        -------
        res: A scalar
            Recall score.

        """
        tp, tp_fn, _ = MeasureAtK.compute(self, gt_pos, pd_rank, **kwargs)
        return tp / tp_fn


class FMeasure(MeasureAtK):
    """F-measure@K.

    Parameters
    ----------
    k: int or list, optional, default: -1 (all)
        The number of items in the top@k list.
        If None, all items will be considered.

    """

    def __init__(self, k=-1):
        super().__init__(name="F1@{}".format(k), k=k)

    def compute(self, gt_pos, pd_rank, **kwargs):
        """Compute F-Measure.

        Parameters
        ----------
        gt_pos: Numpy array
            Binary vector of positive items.

        pd_rank: Numpy array
            Item ranking prediction.

        **kwargs: For compatibility

        Returns
        -------
        res: A scalar
            F-Measure score.

        """
        tp, tp_fn, tp_fp = MeasureAtK.compute(self, gt_pos, pd_rank, **kwargs)

        prec = tp / tp_fp
        rec = tp / tp_fn
        if (prec + rec) > 0:
            f1 = 2 * (prec * rec) / (prec + rec)
        else:
            f1 = 0

        return f1


class AUC(RankingMetric):
    """Area Under the ROC Curve (AUC).

    References
    ----------
    https://arxiv.org/ftp/arxiv/papers/1205/1205.2618.pdf

    """

    def __init__(self):
        RankingMetric.__init__(self, name="AUC")

    def compute(self, pd_scores, gt_pos, gt_neg=None, **kwargs):
        """Compute Area Under the ROC Curve (AUC).

        Parameters
        ----------
        pd_scores: Numpy array
            Prediction scores for items.

        gt_pos: Numpy array
            Binary vector of positive items.

        gt_neg: Numpy array, optional
            Binary vector of negative items.
            If None, negation of gt_pos will be used.

        **kwargs: For compatibility

        Returns
        -------
        res: A scalar
            AUC score.

        """
        if gt_neg is None:
            gt_neg = np.logical_not(gt_pos)

        pos_scores = pd_scores[gt_pos.astype(np.bool)]
        neg_scores = pd_scores[gt_neg.astype(np.bool)]
        ui_scores = np.repeat(pos_scores, len(neg_scores))
        uj_scores = np.tile(neg_scores, len(pos_scores))

        return (ui_scores > uj_scores).sum() / len(uj_scores)


class MAP(RankingMetric):
    """Mean Average Precision (MAP).

    References
    ----------
    https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Mean_average_precision

    """

    def __init__(self):
        RankingMetric.__init__(self, name="MAP")

    def compute(self, pd_scores, gt_pos, **kwargs):
        """Compute Average Precision.

        Parameters
        ----------
        pd_scores: Numpy array
            Prediction scores for items.

        gt_pos: Numpy array
            Binary vector of positive items.

        **kwargs: For compatibility

        Returns
        -------
        res: A scalar
            AP score.

        """
        relevant = gt_pos.astype(np.bool)
        rank = rankdata(-pd_scores, "max")[relevant]
        L = rankdata(-pd_scores[relevant], "max")
        ans = (L / rank).mean()

        return ans



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from libcpp.vector cimport vector

import numpy as np
import scipy.sparse as sp

ctypedef vector[vector[double]] Mat
ctypedef vector[double] dVec
ctypedef vector[int] iVec

cdef extern from "cpp_c2pf.h":
    void c2pf_cpp(Mat &X, Mat &C, int &g, Mat &G_s, Mat &G_r, Mat &L_s, Mat &L_r, Mat &L2_s, Mat &L2_r, Mat &L3_s, Mat &L3_r, dVec &T3_r, dVec &c_sum_c, dVec &util_sum, int maxiter, double at, double bt)
    void c2pf_cpp2(Mat &X, Mat &C, int &g, Mat &G_s, Mat &G_r, Mat &L_s, Mat &L_r, Mat &L2_s, Mat &L2_r, Mat &L3_s, Mat &L3_r, dVec &T3_r, dVec &c_sum_c, dVec &util_sum, int maxiter,double at,double bt)
    void tc2pf_cpp(Mat &X, Mat &C, int &g, Mat &G_s, Mat &G_r, Mat &L_s, Mat &L_r, Mat &L3_s, Mat &L3_r, dVec &T2_r, dVec &c_sum_c, dVec &util_sum, int maxiter, double at, double bt)
    void rc2pf_cpp(Mat &X, Mat &C, int &g, Mat &G_s, Mat &G_r, Mat &L2_s, Mat &L2_r, Mat L3_s, Mat L3_r, dVec &T3_r, dVec &c_sum_c, dVec &util_sum, int maxiter, double at, double bt)
    
#C2PF
cpdef c2pf(Mat &X, int n_X, int d_X, C,int n_C, int d_C, int &k, int &iter_max, init_param = None):
    
    n   = n_X
    d   = d_X
    d2  = d_C

    #Hyper parameter setting
    a  = 0.3
    a_ = 0.3
    a1 = 6.
    a1_t = 3.
    c  = 0.3
    c_ = 0.3
    c1 = 5.
    c1_t = 2.
    e  = 0.3
    a_t = 100.
    b_t = 100./0.5
    b_ = 1
    d_ = 1
    
    #Declare C++ variables
    cdef:
        Mat G_s, G_r, L_s, L_r, L2_s, L2_r, L3_s, L3_r
    
    
    
    #Parameters Initialization
    if init_param['G_s'] is None:
        G_s = np.random.gamma(100,scale=0.3/100,size=n*k).reshape(n,k)
    else:
        G_s = init_param['G_s']

    #G_s = sp.csc_matrix(G_s,dtype=np.float64)
  
    # rate gamma_uk matrix
    if init_param['G_r'] is None:
        G_r = np.random.gamma(100,scale=0.3/100, size=n*k).reshape(n,k)
    else:
        G_r = init_param['G_r']
    #G_r = sp.csc_matrix(G_r,dtype=np.float64)
  
    ## shape lamda_ik matrix (dgCMatrix)
    if init_param['L_s'] is None:
        L_s = np.random.gamma(100,scale=0.3/100, size=d*k).reshape(d,k)
                
    else:
        L_s = init_param['L_s']
    #L_s = sp.csc_matrix(L_s,dtype=np.float64)
  
    #rate lamda_ik matrix (dgCMatrix)
    if init_param['L_r'] is None: 
        L_r = np.random.gamma(100,scale=0.3/100, size=d*k).reshape(d,k)
    else:
        L_r = init_param['L_r']
    #L_r = sp.csc_matrix(L_r,dtype=np.float64)
    
    if init_param['L2_s'] is None:
        L2_s = np.random.gamma(100,scale=0.3/100, size=d2*k).reshape(d2,k)
    else:
        L2_s = init_param['L2_s']
    #L_s = sp.csc_matrix(L_s,dtype=np.float64)
  
    #rate lamda_ik matrix (dgCMatrix)
    if init_param['L2_r'] is None: 
        L2_r = np.random.gamma(100,scale=0.3/100, size=d2*k).reshape(d2,k)
    else:
        L2_r = init_param['L2_r']
    #L2_r = sp.csc_matrix(L_r,dtype=np.float64)
    #shape kappa_ij matrix
    if init_param['L3_s'] is None:
        tmp = np.copy(C)
        tmp[:,2] = np.random.gamma(100,scale=0.5/100, size= C.shape[0])
        L3_s = np.copy(tmp)
        del(tmp)
    else:
        L3_s = init_param['L3_s']

    ## rate kappa_ij matrix (dgCMatrix)
    if init_param['L3_r'] is None: 
        tmp = np.copy(C)
        tmp[:,2] = np.random.gamma(100,scale=0.5/100, size= C.shape[0])
        L3_r = np.copy(tmp)
        del(tmp)
    else:
        L3_r = init_param['L3_r']

    T3_r = np.repeat(1.0,d2)
    col_sum_c = np.repeat(1,d)
    spC = sp.csc_matrix((C[:,2], (C[:,0], C[:,1])), shape=(d, d2))
    util_sum = spC.sum(axis = 0).A1
    del(spC)

    print('Learning...')    
    #c2pf_cpp(X, C, k, G_s, G_r, L_s, L_r, L2_s, L2_r, L3_s, L3_r, T3_r, col_sum_c, util_sum, iter_max,2.,5.)
    c2pf_cpp(X, C, k, G_s, G_r, L_s, L_r, L2_s, L2_r, L3_s, L3_r, T3_r, col_sum_c, util_sum, iter_max,1e15,1e15)
    c2pf_cpp(X, C, k, G_s, G_r, L_s, L_r, L2_s, L2_r, L3_s, L3_r, T3_r, col_sum_c, util_sum, int(0.2*iter_max),2.,5.)
    print('Learning completed!')
    
    M3 = sp.csc_matrix((np.array(L3_s)[:,2]/np.array(L3_r)[:,2], (np.array(L3_s)[:,0], np.array(L3_s)[:,1])), shape=(d, d2))
    Q = M3*(np.array(L2_s)/np.array(L2_r))
    
    res = {
        "Z": np.array(G_s) / np.array(G_r),
        "W": np.array(L_s) / np.array(L_r),
        "ll": None, 
        "Q": Q,
        "G_s": np.array(G_s),
        "G_r": np.array(G_r),
        "L_s": np.array(L_s),
        "L_r": np.array(L_r),
        "L2_s": np.array(L2_s),
        "L2_r": np.array(L2_r),
        "L3_s": np.array(L3_s),
        "L3_r": np.array(L3_r),
    }
    
    return res
    
  
  
#tied-C2PF    
cpdef t_c2pf(Mat &X, int n_X, int d_X, C,int n_C, int d_C, int &k, int &iter_max, init_param = None):
    
    n   = n_X
    d   = d_X
    d2  = d_C

    #Hyper parameter setting
    a  = 0.3
    a_ = 0.3
    c  = 0.3
    c_ = 0.3
    e  = 0.3
    a_t = 100
    b_t = 100/0.5
    b_ = 1
    d_ = 1
    
    #Declare C++ variables
    cdef:
        Mat G_s, G_r, L_s, L_r, L3_s, L3_r
    
    
    
    #Parameters Initialization
    if init_param['G_s'] is None:
        G_s = np.random.gamma(100,scale=0.3/100,size=n*k).reshape(n,k)
    else:
        G_s = init_param['G_s']

    #G_s = sp.csc_matrix(G_s,dtype=np.float64)
  
    # rate gamma_uk matrix
    if init_param['G_r'] is None:
        G_r = np.random.gamma(100,scale=0.3/100, size=n*k).reshape(n,k)
    else:
        G_r = init_param['G_r']
    #G_r = sp.csc_matrix(G_r,dtype=np.float64)
  
    ## shape lamda_ik matrix (dgCMatrix)
    if init_param['L_s'] is None:
        L_s = np.random.gamma(100,scale=0.3/100, size=d*k).reshape(d,k)
                
    else:
        L_s = init_param['L_s']
    #L_s = sp.csc_matrix(L_s,dtype=np.float64)
  
    #rate lamda_ik matrix (dgCMatrix)
    if init_param['L_r'] is None: 
        L_r = np.random.gamma(100,scale=0.3/100, size=d*k).reshape(d,k)
    else:
        L_r = init_param['L_r']
    #L_r = sp.csc_matrix(L_r,dtype=np.float64)
    
    #shape kappa_ij matrix
    if init_param['L3_s'] is None:
        tmp = np.copy(C)
        tmp[:,2] = np.random.gamma(100,scale=0.5/100, size= C.shape[0])
        L3_s = tmp
        del(tmp)
    else:
        L3_s = init_param['L3_s']

    ## rate kappa_ij matrix (dgCMatrix)
    if init_param['L3_r'] is None: 
        tmp = np.copy(C)
        tmp[:,2] = np.random.gamma(100,scale=0.5/100, size= C.shape[0])
        L3_r = tmp
        del(tmp)
    else:
        L3_r = init_param['L3_r']



    T2_r = np.repeat(1.0,d2)
    col_sum_c = np.repeat(1,d)
    spC = sp.csc_matrix((C[:,2], (C[:,0], C[:,1])), shape=(d, d2))
    util_sum = spC.sum(axis = 0).A1
    del(spC)
    
    print('Learning...')
    tc2pf_cpp(X, C, k, G_s, G_r, L_s, L_r, L3_s, L3_r, T2_r, col_sum_c, util_sum, iter_max,1e15,1e15)
    tc2pf_cpp(X, C, k, G_s, G_r, L_s, L_r, L3_s, L3_r, T2_r, col_sum_c, util_sum, int(0.2*iter_max),2.,4.)
    print('Learning completed!')
    
    
    M3 = sp.csc_matrix((np.array(L3_s)[:,2]/np.array(L3_r)[:,2], (np.array(L3_s)[:,0], np.array(L3_s)[:,1])), shape=(d, d2))
    Q = M3*(np.array(L_s)/np.array(L_r))
    
    res = {
        "Z": np.array(G_s) / np.array(G_r),
        "W": np.array(L_s) / np.array(L_r),
        "ll": None, 
        "Q": Q,
        "G_s": np.array(G_s),
        "G_r": np.array(G_r),
        "L_s": np.array(L_s),
        "L_r": np.array(L_r),
        "L2_s": None,
        "L2_r": None,
        "L3_s": np.array(L3_s),
        "L3_r": np.array(L3_r),
    }

    return res  
  
  

#reduced-C2PF
cpdef r_c2pf(Mat &X, int n_X, int d_X, C,int n_C, int d_C, int &k, int &iter_max, init_param = None):
    
    n   = n_X
    d   = d_X
    d2  = d_C

    #Hyper parameter setting
    a  = 0.3
    a_ = 0.3
    a1 = 4.
    c  = 0.3
    c_ = 0.3
    c1 = 3.
    e  = 0.3
    a_t = 100.
    b_t = 100./0.5
    b_ = 1
    d_ = 1
    
    #Declare C++ variables
    cdef:
        Mat G_s, G_r, L2_s, L2_r, L3_s, L3_r
    
    #Parameters Initialization
    if init_param['G_s'] is None:
        G_s = np.random.gamma(100,scale=0.3/100,size=n*k).reshape(n,k)
    else:
        G_s = init_param['G_s']

    #G_s = sp.csc_matrix(G_s,dtype=np.float64)
  
    # rate gamma_uk matrix
    if init_param['G_r'] is None:
        G_r = np.random.gamma(100,scale=0.3/100, size=n*k).reshape(n,k)
    else:
        G_r = init_param['G_r']
    #G_r = sp.csc_matrix(G_r,dtype=np.float64)
    
    if init_param['L2_s'] is None:
        L2_s = np.random.gamma(100,scale=0.3/100, size=d2*k).reshape(d2,k)
    else:
        L2_s = init_param['L2_s']
    #L_s = sp.csc_matrix(L_s,dtype=np.float64)
  
    #rate lamda_ik matrix (dgCMatrix)
    if init_param['L2_r'] is None: 
        L2_r = np.random.gamma(100,scale=0.3/100, size=d2*k).reshape(d2,k)
    else:
        L2_r = init_param['L2_r']
    #L2_r = sp.csc_matrix(L_r,dtype=np.float64)
    #shape kappa_ij matrix
    if init_param['L3_s'] is None: 
        L3_s = np.copy(C)
        L3_s[:,2] = np.random.gamma(100,scale=0.5/100, size= C.shape[0])
    else:
        L3_s = init_param['L3_s']

    ## rate kappa_ij matrix (dgCMatrix)
    if init_param['L3_r'] is None: 
        L3_r = np.copy(C)
        L3_r[:,2] = np.random.gamma(100,scale=0.5/100, size= C.shape[0])
    else:
        L3_r = init_param['L3_r']

    T3_r = np.repeat(1.0,d2)
    col_sum_c = np.repeat(1,d)
    spC = sp.csc_matrix((C[:,2], (C[:,0], C[:,1])), shape=(d, d2))
    util_sum = spC.sum(axis = 0).A1
    del(spC)
    
    print('Learning...')
    rc2pf_cpp(X, C, k, G_s, G_r, L2_s, L2_r, L3_s, L3_r, T3_r, col_sum_c, util_sum, iter_max,1e15,1e15) 
    rc2pf_cpp(X, C, k, G_s, G_r, L2_s, L2_r, L3_s, L3_r, T3_r, col_sum_c, util_sum, int(0.2*iter_max),2.,4.) 
    print('Learning completed!')
    
    M3 = sp.csc_matrix((np.array(L3_s)[:,2]/np.array(L3_r)[:,2], (np.array(L3_s)[:,0], np.array(L3_s)[:,1])), shape=(d, d2))
    Q = M3*(np.array(L2_s)/np.array(L2_r))
    
    res = {
        "Z": np.array(G_s) / np.array(G_r),
        "W": Q,
        "ll": None, 
        "Q": Q,
        "G_s": np.array(G_s),
        "G_r": np.array(G_r),
        "L_s": None,
        "L_r": None,
        "L2_s": np.array(L2_s),
        "L2_r": np.array(L2_r),
        "L3_s": np.array(L3_s),
        "L3_r": np.array(L3_r),
    }

    return res



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
This dataset is mostly from the paper 'Collaborative topic modeling for recommending scientific articles'
[Wang and Blei - KDD 2011].  It was further collected, named `citeulike-a`, and used in the paper
'Collaborative Topic Regression with Social Regularization' [Wang, Chen and Li - IJCAI 2013].

Link to the data: http://www.wanghao.in/CDL.htm
"""

from typing import List

from ..utils import cache
from ..data import Reader


def load_feedback(reader: Reader = None) -> List:
    """Load the implicit feedback between users and items

    Parameters
    ----------
    reader: `obj:cornac.data.Reader`, default: None
        Reader object used to read the data.

    Returns
    -------
    data: array-like
        Data in the form of a list of tuples (user, item, 1).

    """
    fpath = cache(url='https://static.preferred.ai/cornac/datasets/citeulike/users.zip',
                  relative_path='citeulike/users.dat', unzip=True)
    reader = Reader() if reader is None else reader
    return reader.read(fpath, fmt='UI', sep=' ', id_inline=True)


def load_text():
    """Load item texts including tile and abstract joined together into one document per item.

    Returns
    -------
    texts: List
        List of text documents, one per item.

    ids: List
        List of item ids aligned with indices in `texts`.
    """
    import csv

    texts, ids = [], []
    fpath = cache(url='https://static.preferred.ai/cornac/datasets/citeulike/text.zip',
                  relative_path='citeulike/raw-data.csv', unzip=True)
    with open(fpath, 'r', encoding='utf-8', errors='ignore') as f:
        next(f)
        for row in csv.reader(f, delimiter=',', quotechar='"'):
            ids.append(row[0])
            texts.append(row[3] + '. ' + row[4])

    return texts, ids



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

# cython: language_level=3

cimport cython
from cython cimport floating, integral
from cython.parallel import parallel, prange

import numpy as np
cimport numpy as np

from ..bpr.recom_bpr import BPR
from ..bpr.recom_bpr cimport RNGVector, has_non_zero


cdef extern from "../bpr/recom_bpr.h" namespace "recom_bpr" nogil:
    cdef int get_thread_num()


class MMMF(BPR):
    """Maximum Margin Matrix Factorization.
    This implements MF model optimized for the Soft Margin (Hinge) Ranking Loss, \
    using SGD as similar to BPR model.

    Parameters
    ----------
    k: int, optional, default: 10
        The dimension of the latent factors.

    max_iter: int, optional, default: 100
        Maximum number of iterations or the number of epochs for SGD.

    learning_rate: float, optional, default: 0.001
        The learning rate for SGD.

    lambda_reg: float, optional, default: 0.001
        The regularization hyper-parameter.

    num_threads: int, optional, default: 0
        Number of parallel threads for training. If num_threads=0, all CPU cores will be utilized.
        If seed is not None, num_threads=1 to remove randomness from parallelization.

    trainable: boolean, optional, default: True
        When False, the model will not be re-trained, and input of pre-trained parameters are required.

    verbose: boolean, optional, default: True
        When True, some running logs are displayed.

    init_params: dictionary, optional, default: None
        Initial parameters, e.g., init_params = {'U': user_factors, 'V': item_factors, 'Bi': item_biases}

    seed: int, optional, default: None
        Random seed for weight initialization.
        If specified, training will take longer because of single-thread (no parallelization).

    References
    ----------
    * Weimer, M., Karatzoglou, A., & Smola, A. (2008). Improving maximum margin matrix factorization. Machine Learning, 72(3), 263-276.
    """

    def __init__(
        self, 
        name='MMMF', 
        k=10, 
        max_iter=100, 
        learning_rate=0.001, 
        lambda_reg=0.01,
        num_threads=0, 
        trainable=True, 
        verbose=False, 
        init_params=None, 
        seed=None,
    ):
        super().__init__(
            name=name, 
            k=k, 
            max_iter=max_iter, 
            learning_rate=learning_rate, 
            lambda_reg=lambda_reg, 
            num_threads=num_threads, 
            trainable=trainable, 
            verbose=verbose, 
            init_params=init_params, 
            seed=seed,
        )

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _fit_sgd(self, RNGVector rng_pos, RNGVector rng_neg, int num_threads,
                 integral[:] user_ids, integral[:] item_ids, 
                 integral[:] neg_item_ids, integral[:] indptr,
                 floating[:, :] U, floating[:, :] V, floating[:] B):
        """Fit the model parameters (U, V, B) with SGD
        """
        cdef:
            long num_samples = len(user_ids), s, i_index, j_index, correct = 0, skipped = 0
            long num_items = self.train_set.num_items
            integral f, i_id, j_id, thread_id
            floating z, score, temp

            floating lr = self.learning_rate
            floating reg = self.lambda_reg
            int factors = self.k

            floating * user
            floating * item_i
            floating * item_j

        with nogil, parallel(num_threads=num_threads):
            thread_id = get_thread_num()

            for s in prange(num_samples, schedule='guided'):
                i_index = rng_pos.generate(thread_id)
                i_id = item_ids[i_index]
                j_index = rng_neg.generate(thread_id)
                j_id = neg_item_ids[j_index]

                # if the user has liked the item j, skip this for now
                if has_non_zero(indptr, item_ids, user_ids[i_index], j_id):
                    skipped += 1
                    continue

                # get pointers to the relevant factors
                user, item_i, item_j = &U[user_ids[i_index], 0], &V[i_id, 0], &V[j_id, 0]

                # compute the score
                score = B[i_id] - B[j_id]
                for f in range(factors):
                    score = score + user[f] * (item_i[f] - item_j[f])

                if score > 0:
                    correct += 1
                    continue

                # update the factors via sgd.
                for f in range(factors):
                    temp = user[f]
                    user[f] += lr * (item_i[f] - item_j[f] - reg * user[f])
                    item_i[f] += lr * (temp - reg * item_i[f])
                    item_j[f] += lr * (-temp - reg * item_j[f])

                # update item biases
                B[i_id] += lr * (1 - reg * B[i_id])
                B[j_id] += lr * (-1 - reg * B[j_id])

        return correct, skipped